import os
from typing import Optional
import xml.etree.ElementTree as etree

from .xmlns import dataset_as_xml
from .xmlns import investigation_as_xml
from .messaging import IcatMessagingClient


class IcatMetadataClient:
    """Client for storing dataset metadata in ICAT."""

    def __init__(
        self,
        queue_urls: list,
        queue_name: str = "icatIngest",
        monitor_port: Optional[int] = None,
        timeout: Optional[float] = None,
    ):
        self._client = IcatMessagingClient(
            queue_urls, queue_name, monitor_port=monitor_port, timeout=timeout
        )

    def send_metadata(
        self,
        beamline: str,
        proposal: str,
        dataset: str,
        path: str,
        metadata: dict,
    ):
        root = dataset_as_xml(
            beamline=beamline,
            proposal=proposal,
            dataset=dataset,
            path=path,
            metadata=metadata,
        )
        self._client.send(etree.tostring(root))

    def store_metadata(
        self,
        filename: str,
        beamline: str,
        proposal: str,
        dataset: str,
        path: str,
        metadata: dict,
    ):
        root = dataset_as_xml(
            beamline=beamline,
            proposal=proposal,
            dataset=dataset,
            path=path,
            metadata=metadata,
        )
        filename, ext = os.path.splitext(filename)
        if not ext:
            ext = ".xml"
        filename += ext
        os.makedirs(os.path.dirname(filename), exist_ok=True)
        with open(filename, "wb") as f:
            f.write(etree.tostring(root))

    def send_metadata_from_file(self, filename: str):
        filename, ext = os.path.splitext(filename)
        if not ext:
            ext = ".xml"
        filename += ext
        with open(filename, "rb") as f:
            payload = f.read()
        self._client.send(payload)

    def start_investigation(self, beamline: str, proposal: str, start_datetime=None):
        root = investigation_as_xml(
            beamline=beamline, proposal=proposal, start_datetime=start_datetime
        )
        self._client.send(etree.tostring(root))

    def check_health(self):
        """Raises an exception when not healthy"""
        self._client.check_health()



import pytest

from ..client.archive import StatusLevel, StatusType


def test_send_archive_status(icat_archive_client):
    client, messages = icat_archive_client
    client.check_health()
    client.send_archive_status(
        dataset_id=12345,
        type=StatusType.ARCHIVING,
        level=StatusLevel.INFO,
        message="dataset archived",
    )
    messages.get(timeout=10)
    assert messages.empty()


def test_send_archive_status_raises_error_if_dataset_id_missing(icat_archive_client):
    client, messages = icat_archive_client
    with pytest.raises(AssertionError, match="ICAT requires the datasetId"):
        client.send_archive_status(
            dataset_id=None,
            type=StatusType.ARCHIVING,
            level=StatusLevel.INFO,
            message="dataset archived",
        )


def test_send_archive_status_raises_error_if_type_missing(icat_archive_client):
    client, messages = icat_archive_client
    with pytest.raises(AssertionError, match="ICAT requires the type"):
        client.send_archive_status(
            dataset_id=123,
            type=None,
            level=StatusLevel.INFO,
            message="dataset archived",
        )


def test_send_archive_status_raises_error_if_level_missing(icat_archive_client):
    client, messages = icat_archive_client
    with pytest.raises(AssertionError, match="ICAT requires the level"):
        client.send_archive_status(
            dataset_id=123,
            type=StatusType.RESTORATION,
            level=None,
            message="dataset restored",
        )



"""
Nexus structure from ICAT metadata
"""

from typing import Optional
from collections.abc import Iterable
import logging

from .definitions import load_icat_fields
from .definitions import IcatFieldGroup


logger = logging.getLogger(__name__)


def as_nxtype(value, nxtype):
    """Nexus convert to the Nexus type as defined in the ICAT definitions"""
    if nxtype is None:
        return value
    elif nxtype == "NX_CHAR":
        return value  # TODO: the ICAT definitions abuse NX_CHAR
        if isinstance(value, str):
            return value
        elif isinstance(value, bytes):
            return str(value)
        elif isinstance(value, Iterable):
            return [str(s) for s in value]
        else:
            return str(value)
    elif nxtype == "NX_DATE_TIME":
        if not isinstance(value, str):
            value = value.isoformat()
        return str(value)
    elif nxtype == "NX_INT":
        return int(value)
    elif nxtype == "NX_FLOAT":
        return float(value)
    else:
        return value


def create_nxtreedict(metadata: dict, icat_fields: Optional[IcatFieldGroup] = None):
    if icat_fields is None:
        icat_fields = load_icat_fields()
    nxtreedict = dict()
    for field_name, field_value in metadata.items():
        field = icat_fields.get_field_with_field_name(field_name)
        if field is None:
            logger.warning(f"{field_name} not a valid ICAT field")
            continue

        # Prepare parent groups
        group = icat_fields
        adict = nxtreedict
        if group.info.NX_class:
            adict["@NX_class"] = group.info.NX_class
        for name in field.parent:
            group = group[name]
            adict = adict.setdefault(name, dict())
            if group.info.NX_class:
                adict["@NX_class"] = group.info.NX_class

        # Add value
        adict[field.name] = as_nxtype(field_value, field.nxtype)
        if field.units:
            adict[f"{field.name}@units"] = field.units
    return nxtreedict



import base64
import requests
from typing import Optional

from stompest.config import StompConfig
from stompest.protocol import StompSpec
from stompest.protocol import StompSession
from stompest.sync import Stomp

from ..utils.url import normalize_url


class IcatMessagingClient:
    """Client for the ICAT message broker.

    The message broker is currently ActiveMQ, a message
    broker using the STOMP protocol. It also has a REST
    server for monitoring the broker status.
    """

    DEFAULT_SCHEME = "tcp"
    DEFAULT_PORT = 61613
    MONITOR_SCHEME = "http"
    DEFAULT_MONITOR_PORT = 8778
    MONITOR_USER = "user"
    MONITOR_PWD = "user"

    def __init__(
        self,
        queue_urls: list,
        queue_name: str,
        monitor_port: Optional[int] = None,
        timeout: Optional[float] = None,
    ):
        urls = [
            normalize_url(
                url, default_scheme=self.DEFAULT_SCHEME, default_port=self.DEFAULT_PORT
            )
            for url in queue_urls
        ]
        failover = ",".join(urls)
        url = f"failover:({failover})?maxReconnectAttempts=3,initialReconnectDelay=250,maxReconnectDelay=1000"
        self.__max_version = StompSpec.VERSION_1_1
        self.__client = Stomp(StompConfig(url, version=self.__max_version))
        self.socket_timeout = None
        if timeout is None:
            timeout = 1
        self.connect_timeout = timeout

        self.__send_destination = "/queue/" + queue_name
        self.__send_headers = {
            "persistent": "true",
            StompSpec.ACK_HEADER: StompSpec.ACK_CLIENT_INDIVIDUAL,
        }

        if not monitor_port:
            monitor_port = self.DEFAULT_MONITOR_PORT
        self.__consumer_count_url = f"{self.MONITOR_SCHEME}://{{host}}:{monitor_port}/api/jolokia/read/org.apache.activemq:type=Broker,brokerName=metadata,destinationType=Queue,destinationName={queue_name}/ConsumerCount"
        self.__jolokia_headers = {
            "Authorization": b"Basic "
            + base64.b64encode(f"{self.MONITOR_USER}:{self.MONITOR_PWD}".encode())
        }

    def close(self):
        self.__client.close()

    @property
    def _connected_client(self):
        if self.__client.session.state != StompSession.CONNECTED:
            self.__client.connect(
                versions=[self.__max_version],
                connectTimeout=self.socket_timeout,
                connectedTimeout=self.connect_timeout,
            )
        return self.__client

    @property
    def _host(self):
        return self._connected_client._transport.host

    def send(self, data: bytes):
        self._connected_client.send(
            self.__send_destination, body=data, headers=self.__send_headers
        )

    @property
    def _consumer_count(self):
        url = self.__consumer_count_url.format(host=self._host)
        response = requests.get(url, headers=self.__jolokia_headers)
        if not response.ok:
            raise RuntimeError(
                response, "Failed to retrieve the ActiveMQ consumer count"
            )
        response = response.json()
        return response["value"]

    def check_health(self):
        """Raises an exception when:
        - not connected
        - no message consumers
        """
        state = self._connected_client.session.state
        if state != StompSession.CONNECTED:
            raise RuntimeError(
                "The connection with the message broker is " + str(state).upper()
            )
        if self._consumer_count < 1:
            raise RuntimeError("The message broker has no consumers")



import sys
import logging
import socketserver


def basic_config(logger=None, level=None, format=None):
    """
    :param logger: root logger when not provided
    :param level: logger log level
    :param str format:
    """
    if logger is None:
        logger = logging.getLogger()
    if level is not None:
        logger.setLevel(level)
    if format:
        formatter = logging.Formatter(format)
    else:
        formatter = None

    class StdOutFilter(logging.Filter):
        def filter(self, record):
            return record.levelno < logging.WARNING

    class StdErrFilter(logging.Filter):
        def filter(self, record):
            return record.levelno >= logging.WARNING

    h = logging.StreamHandler(sys.stdout)
    h.addFilter(StdOutFilter())
    h.setLevel(logging.DEBUG)
    if formatter is not None:
        h.setFormatter(formatter)
    logger.addHandler(h)

    h = logging.StreamHandler(sys.stderr)
    h.addFilter(StdErrFilter())
    h.setLevel(logging.WARNING)
    if formatter is not None:
        h.setFormatter(formatter)
    logger.addHandler(h)


class ReuseAddrTCPServer(socketserver.TCPServer):
    allow_reuse_address = True



import pytest
from datetime import datetime
from ..client.serialize import serialize_metadata


def test_icat_serialize_valid_data():
    assert serialize_metadata(None) is None
    assert serialize_metadata("string") == "string"
    assert serialize_metadata(b"string") == "string"
    assert serialize_metadata(123.456) == "123.456"

    assert serialize_metadata([]) == ""
    assert serialize_metadata([None]) == ""
    assert serialize_metadata([1, 2]) == "1,2"
    assert serialize_metadata([[1, 2], [3, 4]]) == "1,2 3,4"

    assert serialize_metadata({}) == {}
    assert serialize_metadata({"key": None}) == {}
    assert serialize_metadata({"key": "string"}) == {"key": "string"}
    assert serialize_metadata({"key": b"string"}) == {"key": "string"}
    assert serialize_metadata({"key": 123.456}) == {"key": "123.456"}
    assert serialize_metadata({"key": []}) == {"key": ""}
    assert serialize_metadata({"key": [None]}) == {"key": ""}
    assert serialize_metadata({"key": [1, 2]}) == {"key": "1,2"}
    assert serialize_metadata({"key": [[1, 2], [3, 4]]}) == {"key": "1,2 3,4"}

    now = datetime.now()
    assert serialize_metadata({"key": now}) == {"key": now.isoformat()}


def test_icat_serialize_invalid_data():
    invalid_data = ([{}], [[[1, 2], [3, 4]], [[5, 6], [7, 8]]])
    for data in invalid_data:
        with pytest.raises((TypeError, ValueError)):
            print(serialize_metadata(data))



from collections.abc import Mapping
import numpy


def deep_compare(adict: Mapping, expected: Mapping):
    """using logic of deep update to compare two dictionaries"""
    stack = [(adict, expected)]
    while stack:
        adict, expected = stack.pop(0)
        existing_keys = set(adict.keys())
        expected_keys = set(expected.keys())
        missing_keys = expected_keys - existing_keys
        unexpected_keys = existing_keys - expected_keys
        if missing_keys:
            raise AssertionError(f"missing keys: {missing_keys}")
        if unexpected_keys:
            raise AssertionError(f"unexpected keys: {unexpected_keys}")
        for k, v in expected.items():
            if isinstance(v, Mapping):
                stack.append((adict[k], v))
            elif isinstance(v, numpy.ndarray) and v.size > 1:
                assert adict[k].shape == v.shape
                assert adict[k].dtype == v.dtype
                if adict[k].dtype != object:
                    assert all(
                        numpy.isnan(adict[k].flatten()) == numpy.isnan(v.flatten())
                    )
                    mask = numpy.logical_not(numpy.isnan(v.flatten()))
                    assert all((adict[k].flatten() == v.flatten())[mask])
                else:
                    assert all(adict[k].flatten() == v.flatten())
            elif isinstance(v, (list, tuple)):
                assert list(adict[k]) == list(v)
            else:
                assert adict[k] == v



import sys
import json
import argparse
from configparser import RawConfigParser
from typing import Any, Tuple
from ..client.main import IcatClient


def main(argv=None):
    if argv is None:
        argv = sys.argv

    parser = argparse.ArgumentParser(description="Register processed data with ICAT")
    add_store_parameters(parser)
    args = parser.parse_args(argv[1:])
    apply_store_parameters(args)

    client = IcatClient(metadata_urls=args.metadata_urls)
    client.store_dataset(
        beamline=args.beamline,
        proposal=args.proposal,
        dataset=args.dataset,
        path=args.path,
        metadata=args.metadata,
    )


def add_store_parameters(parser):

    parser.add_argument("--beamline", required=True, help="Beamline name")

    parser.add_argument("--proposal", required=True, help="Proposal name")

    parser.add_argument("--dataset", required=True, help="Dataset name")

    parser.add_argument("--sample", required=True, help="Sample name")

    parser.add_argument(
        "--path", required=True, help="Directory of the data to be registered"
    )

    parser.add_argument(
        "--metadatafile",
        type=argparse.FileType("r"),
        help="ICAT metadata parameters file",
    )

    parser.add_argument(
        "-p",
        "--parameter",
        dest="parameters",
        action="append",
        metavar="NAME=VALUE",
        help="ICAT metadata names and values (overwrite metadata from file)",
    )

    parser.add_argument(
        "--queue",
        dest="metadata_urls",
        action="append",
        help="ActiveMQ queue URLS. Example: bcu-mq-01.esrf.fr:61613",
        default=[],
    )


def apply_store_parameters(args):
    if args.metadatafile:
        parameters = _parameters_from_file()
    else:
        parameters = dict()
    if args.parameters:
        parameters.update(_parse_parameter(s) for s in args.parameters)
    if args.sample:
        parameters["Sample_name"] = args.sample
    args.metadata = parameters

    if not args.metadata_urls:
        args.metadata_urls = ["bcu-mq-01.esrf.fr:61613", "bcu-mq-02.esrf.fr:61613"]


def _parse_parameter(parameter: str) -> Tuple[str, str]:
    name, _, value = parameter.partition("=")
    return name, _parse_value(value)


def _parse_value(value: str) -> Any:
    try:
        return json.loads(value)
    except Exception:
        return value


def _parameters_from_file(path: str) -> dict:
    config = RawConfigParser()
    with open(path, "r") as f:
        config.read_string("[config]\n" + f.read())
        return dict(config["config"].items())


if __name__ == "__main__":
    sys.exit(main())



import json
from enum import Enum
from typing import Optional

from .messaging import IcatMessagingClient


class StatusType(Enum):
    ARCHIVING = "archiving"
    RESTORATION = "restoration"


class StatusLevel(Enum):
    INFO = "info"
    WARNING = "warning"
    ERROR = "error"


class IcatArchiveStatusClient:
    """Client for storing archive and restoration status in ICAT."""

    def __init__(
        self,
        queue_urls: list,
        queue_name: str = "icatArchiveRestoreStatus",
        monitor_port: Optional[int] = None,
        timeout: Optional[float] = None,
    ):
        self._client = IcatMessagingClient(
            queue_urls, queue_name, monitor_port=monitor_port, timeout=timeout
        )

    def send_archive_status(
        self, dataset_id: int, type: StatusType, level: StatusLevel, message: str
    ):
        assert dataset_id, "ICAT requires the datasetId"
        assert type, "ICAT requires the type"
        assert level, "ICAT requires the level"
        root = {
            "datasetId": dataset_id,
            "type": type.value,
            "level": level.value,
            "message": message,
        }
        data = json.dumps(root).encode("utf-8")
        self._client.send(data)

    def check_health(self):
        """Raises an exception when not healthy"""
        self._client.check_health()



import pytest
import xml.etree.ElementTree as etree
from ..concurrency import Empty


def test_start_investigation(icat_metadata_client):
    client, messages = icat_metadata_client
    client.check_health()
    client.start_investigation(proposal="hg123", beamline="id00")
    message = messages.get(timeout=10)

    root = etree.fromstring(message)
    names = {child.tag.replace("{http://www.esrf.fr/icat}", "") for child in root}
    expected = {"startDate", "experiment", "instrument"}
    assert names == expected
    assert messages.empty()


def test_start_bad_investigation(icat_metadata_client):
    client, messages = icat_metadata_client
    client.check_health()
    client.start_investigation(proposal="hg666", beamline="id00")
    with pytest.raises(Empty):
        messages.get(timeout=2)


def test_send_metadata(icat_metadata_client):
    client, messages = icat_metadata_client
    client.send_metadata(
        proposal="hg123",
        beamline="id00",
        dataset="datasetname",
        path="/path-of-dataset",
        metadata={"Sample_name": "samplename", "field1": "value1", "field2": [1, 2, 3]},
    )
    message = messages.get(timeout=10)

    root = etree.fromstring(message)
    names = {child.tag.replace("{http://www.esrf.fr/icat}", "") for child in root}
    expected = {
        "endDate",
        "location",
        "startDate",
        "parameter",
        "sample",
        "investigation",
        "instrument",
        "name",
    }
    assert names == expected
    assert messages.empty()


def test_send_metadata_via_file(icat_metadata_client, tmpdir):
    store_filename = tmpdir / "test.xml"

    client, messages = icat_metadata_client
    client.store_metadata(
        str(store_filename),
        proposal="hg123",
        beamline="id00",
        dataset="datasetname",
        path="/path-of-dataset",
        metadata={"Sample_name": "samplename", "field1": "value1", "field2": [1, 2, 3]},
    )

    with pytest.raises(Empty):
        message = messages.get(timeout=1)

    assert store_filename.exists()

    client.send_metadata_from_file(str(store_filename))

    message = messages.get(timeout=10)

    root = etree.fromstring(message)
    names = {child.tag.replace("{http://www.esrf.fr/icat}", "") for child in root}
    expected = {
        "endDate",
        "location",
        "startDate",
        "parameter",
        "sample",
        "investigation",
        "instrument",
        "name",
    }
    assert names == expected
    assert messages.empty()


def test_send_missing_data(icat_metadata_client):
    client, messages = icat_metadata_client
    with pytest.raises(AssertionError, match="ICAT requires the beamline name"):
        client.send_metadata(
            proposal=None,
            beamline=None,
            dataset=None,
            path=None,
            metadata=None,
        )


def test_send_missing_metadata(icat_metadata_client):
    client, messages = icat_metadata_client
    with pytest.raises(
        AssertionError, match="ICAT metadata field 'Sample_name' is missing"
    ):
        client.send_metadata(
            proposal="hg123",
            beamline="id00",
            dataset="datasetname",
            path="/path-of-dataset",
            metadata={},
        )



import xml.etree.ElementTree as etree
from ..apps import store_raw
from ..apps import store_processed


def test_store_raw(icat_metadata_client, icat_backend):
    _, messages = icat_metadata_client

    argv = [
        "",
        "--beamline",
        "id00",
        "--proposal",
        "hg123",
        "--dataset",
        "datasetname",
        "--path",
        "/path/of/dataset",
        "--sample",
        "samplename",
        "-p",
        "field1=value1",
        "-p",
        "field2=[1, 2, 3]",
        "--queue",
        icat_backend["metadata_urls"][0],
    ]
    store_raw.main(argv)

    message = messages.get(timeout=10)

    root = etree.fromstring(message)
    names = {child.tag.replace("{http://www.esrf.fr/icat}", "") for child in root}
    expected = {
        "endDate",
        "location",
        "startDate",
        "parameter",
        "sample",
        "investigation",
        "instrument",
        "name",
    }
    assert names == expected
    assert messages.empty()


def test_store_processed(icat_metadata_client, icat_backend):
    _, messages = icat_metadata_client

    argv = [
        "",
        "--beamline",
        "id00",
        "--proposal",
        "hg123",
        "--dataset",
        "datasetname",
        "--path",
        "/path/of/processed",
        "--sample",
        "samplename",
        "-p",
        "field1=value1",
        "-p",
        "field2=[1, 2, 3]",
        "--queue",
        icat_backend["metadata_urls"][0],
        "--raw",
        "/path/of/dataset1",
        "--raw",
        "/path/of/dataset2",
    ]
    store_processed.main(argv)

    message = messages.get(timeout=10)

    root = etree.fromstring(message)
    names = {child.tag.replace("{http://www.esrf.fr/icat}", "") for child in root}
    expected = {
        "endDate",
        "location",
        "startDate",
        "parameter",
        "sample",
        "investigation",
        "instrument",
        "name",
    }
    assert names == expected
    assert messages.empty()



import os
import psutil
import pytest
from .misc import eprint
from ...concurrency import wait_process


__all__ = ["cleanup_subprocesses"]


@pytest.fixture
def cleanup_subprocesses():
    yield
    if os.name == "nt":
        current_process = psutil.Process()
        for p in current_process.children():
            wait_terminate(p)


def wait_terminate(process, timeout=10):
    """Try to terminate or kill a process.

    :param process: A process object from `subprocess` or `psutil`, or an PID int
    :param timeout: Timeout to way before using a kill signal
    :raises AssertionError: If the kill fails
    """
    if isinstance(process, int):
        try:
            name = str(process)
            process = psutil.Process(process)
        except Exception:
            # PID is already dead
            return
    else:
        name = repr(" ".join(process.args))
        if process.poll() is not None:
            eprint(f"Process {name} already terminated with code {process.returncode}")
            return
    process.terminate()
    terminated = wait_process(process, 2 * timeout / 3)
    if terminated:
        return
    process.kill()
    terminated = wait_process(process, timeout / 3)
    assert terminated, f"Process {name} doesn't finish: try to kill it..."



from typing import Optional
from collections import OrderedDict


class MaxSizeDict(OrderedDict):
    def __init__(self, *args, maxsize: Optional[int] = None, **kwargs):
        self.__maxsize = maxsize
        super().__init__(*args, **kwargs)
        self.__purge()

    def __setitem__(self, key, value):
        super().__setitem__(key, value)
        self.__purge()

    def __purge(self):
        if self.__maxsize is not None:
            while len(self) > self.__maxsize:
                self.popitem(last=False)



from typing import Any, Callable, Dict, Optional
import gevent

from ...utils.maxsizedict import MaxSizeDict


class CancelQuery(gevent.GreenletExit):
    pass


class QueryPool:
    """Query pool for gevent cooperative calls."""

    def __init__(
        self, timeout: Optional[float] = None, maxqueries: Optional[int] = None
    ):
        """
        :param timeout: The default timeout of a call before returning/raising the previous result.
        :param maxqueries: The maximal number of different queries to store results from.
                           A query can differ in terms of function and/or arguments.
        """
        if not gevent:
            raise RuntimeError("QueryPool requires gevent")
        if timeout is None:
            timeout = 0.1
        self.timeout = timeout
        self.__futures: Dict[tuple, gevent.Greenlet] = dict()
        if maxqueries:
            self.__results = MaxSizeDict(maxsize=maxqueries)
        else:
            self.__results = dict()

    def execute(
        self,
        query: Callable,
        args: Optional[tuple] = tuple(),
        kwargs: Optional[dict] = None,
        timeout: Optional[float] = None,
        default=None,
    ) -> Any:
        """
        :param query:
        :param args: positional arguments
        :param kwargs: named arguments
        :param timeout: the timeout of a call before returning/raising the previous result
        :param default: the default value in case there is no previous result
        :returns: the result of the query or the default value
        :raises: the exception from the query
        """
        if kwargs is None:
            kwargs = dict()
        call_id = query, args, tuple(kwargs.items())
        future = self.__futures.get(call_id)
        if future is None:

            def wrapper():
                try:
                    self.__results[call_id] = False, query(*args, **kwargs)
                except BaseException as e:
                    self.__results[call_id] = True, e
                    raise
                finally:
                    self.__futures.pop(call_id, None)

            future = gevent.Greenlet(wrapper)
            self.__futures[call_id] = future
            future.start()
        if timeout is None:
            timeout = self.timeout
        future.join(timeout=timeout)
        result = self.__results.get(call_id, None)
        if result is None:
            return default
        is_error, result = result
        if is_error:
            raise result
        return result

    def wait(self, timeout=None) -> bool:
        """
        :param timeout:
        :returns: `True` when all queries finished, `False` otherwise
        """
        futures = list(self.__futures.values())
        finished = gevent.joinall(futures, timeout=timeout)
        return len(futures) == len(finished)

    def cancel(self, timeout=None, block=True) -> Optional[bool]:
        """
        :param block:
        :param timeout: only applies when `block=True`
        :returns: `None` when `block=False`, `True` when all queries are
                  cancelled and `False` otherwise
        """
        futures = list(self.__futures.values())
        if not block:
            gevent.killall(futures, exception=CancelQuery, block=False)
            return
        try:
            with gevent.Timeout(timeout) as local_timeout:
                gevent.killall(futures, exception=CancelQuery, timeout=timeout)
        except gevent.Timeout as raised_timeout:
            if raised_timeout is not local_timeout:
                raise
            return False
        return True



import os
from datetime import datetime
from typing import Optional, List
from urllib.parse import urljoin
import requests
import numpy
import logging

from ..concurrency.query_pool import QueryPool
from .interface import DatasetId
from ..utils.maxsizedict import MaxSizeDict
from ..utils.url import normalize_url

logger = logging.getLogger(__name__)


def arg_smallest_positive(arr: numpy.ndarray) -> Optional[int]:
    condition = arr >= 0
    if condition.any():
        return numpy.where(condition, arr, numpy.inf).argmin()


class IcatInvestigationClient:
    """Client for the investigation part of the ICAT+ REST API.

    An "investigation" is a time slot assigned to a particular proposal
    at a particular beamline.

    REST API docs:
    https://icatplus.esrf.fr/api-docs/

    The ICAT+ server project:
    https://gitlab.esrf.fr/icat/icat-plus/-/blob/master/README.md
    """

    DEFAULT_SCHEME = "https"

    def __init__(
        self,
        url: str,
        api_key: str = "elogbook-00000000-0000-0000-0000-000000000000",
        timeout: Optional[float] = None,
    ):
        url = normalize_url(url, default_scheme=self.DEFAULT_SCHEME)

        path = f"dataacquisition/{api_key}/investigation"
        query = (
            "?instrumentName={beamline}&investigationName={proposal}&sortBy=startdate"
        )
        self._investigation_url = urljoin(url, path + query)

        path = f"dataacquisition/{api_key}/dataset"
        query = "?investigationId={investigation_id}"
        self._dataset_url = urljoin(url, path + query)

        self.raise_error = False
        self.__query_pool = QueryPool(timeout=timeout, maxqueries=20)
        self.__investigation_info = MaxSizeDict(maxsize=20)

    @property
    def timeout(self):
        return self.__query_pool.timeout

    @timeout.setter
    def timeout(self, value: Optional[float] = None):
        self.__query_pool.timeout = value

    def _get_with_response_parsing(
        self, url: str, timeout: Optional[float] = None
    ) -> Optional[list]:
        """Return `None` means the information is not available at this moment.
        An empty list means that an error has occured or an actual empty list
        is returned.
        """
        try:
            response = self.__query_pool.execute(
                requests.get, args=(url,), timeout=timeout, default=None
            )
        except requests.exceptions.ReadTimeout:
            return None
        except Exception as e:
            if self.raise_error:
                raise
            logger.exception(e)
            return None
        if response is None:
            return None
        if self.raise_error:
            response.raise_for_status()
        elif not response.ok:
            logger.error("%s: %s", response, response.text)
        if response.ok:
            return response.json()
        else:
            return list()

    def investigation_info(
        self, beamline: str, proposal: str, timeout: Optional[float] = None
    ) -> Optional[dict]:
        investigation_key = beamline, proposal
        ninfo = self.__investigation_info.get(investigation_key)
        if ninfo is not None:
            return ninfo

        # Get all investigations for this proposal and beamline
        url = self._investigation_url.format(beamline=beamline, proposal=proposal)
        investigations = self._get_with_response_parsing(url, timeout=timeout)
        if investigations is None:
            return None  # not available at the moment
        if not investigations:
            return dict()  # error or no investigation found

        # Get the closest investigation which started before "now".
        # If there is no such investigation, get the closest investigation
        # which starts after "time".
        now = datetime.now().astimezone()
        seconds_since_start = numpy.array(
            [
                (now - datetime.fromisoformat(info["startDate"])).total_seconds()
                for info in investigations
            ]
        )
        idx = arg_smallest_positive(seconds_since_start)
        if idx is None:
            idx = arg_smallest_positive(-seconds_since_start)
        info = investigations[idx]

        # Normalize information
        for key in ["parameters", "visitId"]:
            info.pop(key, None)
        ninfo = dict()
        ninfo["proposal"] = info.pop("name", None)
        ninfo["beamline"] = info.pop("instrument", dict()).get("name", None)
        ninfo.update(info)
        ninfo["e-logbook"] = f"https://data.esrf.fr/investigation/{info['id']}/events"
        ninfo[
            "data portal"
        ] = f"https://data.esrf.fr/investigation/{info['id']}/datasets"

        self.__investigation_info[investigation_key] = ninfo
        return ninfo

    def _investigation_id(
        self, beamline: str, proposal: str, timeout: Optional[float] = None
    ) -> Optional[int]:
        info = self.investigation_info(
            beamline=beamline, proposal=proposal, timeout=timeout
        )
        if info is None:
            return None
        return info.get("id", None)

    def registered_dataset_ids(
        self, beamline: str, proposal: str, timeout: Optional[float] = None
    ) -> Optional[List[DatasetId]]:
        investigation_id = self._investigation_id(
            beamline=beamline, proposal=proposal, timeout=timeout
        )
        if investigation_id is None:
            return None
        url = self._dataset_url.format(investigation_id=investigation_id)
        datasets = self._get_with_response_parsing(url, timeout=timeout)
        if datasets is None:
            return None
        return [self._icat_dataset_to_datasetid(dataset) for dataset in datasets]

    @staticmethod
    def _icat_dataset_to_datasetid(dataset: dict) -> DatasetId:
        location = dataset["location"]
        location, name = os.path.split(location)
        while location and not name:
            location, name = os.path.split(location)
        return DatasetId(name=name, path=dataset["location"])



import pytest
import base64


def test_elogbook_message_wrong_category(elogbook_client):
    client, messages = elogbook_client
    with pytest.raises(ValueError):
        client.send_message(
            "mycontent",
            category="wrongcategory",
            beamline="id00",
            proposal="hg123",
            dataset="datasetname",
        )
    assert messages.empty()


def test_elogbook_message(elogbook_client):
    client, messages = elogbook_client
    client.send_message(
        "mycontent",
        category="comment",
        beamline="id00",
        proposal="hg123",
        dataset="datasetname",
    )
    message = messages.get(timeout=10)
    assert message.pop("apikey")
    assert message.pop("creationDate")
    assert message.pop("machine")
    assert message.pop("software").startswith("pyicat-plus")
    expected = {
        "type": "annotation",
        "datasetName": "datasetname",
        "category": "comment",
        "content": [{"format": "plainText", "text": "mycontent"}],
        "investigation": "hg123",
        "instrument": "id00",
    }
    assert message == expected
    assert messages.empty()


def test_ebs_elogbook_message(elogbook_ebs_client):
    client, messages = elogbook_ebs_client

    client.send_message("mycontent", category="error")
    message = messages.get(timeout=10)
    assert message.pop("apikey")
    assert message.pop("creationDate")
    assert message.pop("machine")
    expected = {
        "type": "broadcast",
        "source": "ebs",
        "category": "error",
        "software": "testsoft",
        "tag": [{"name": "testtag"}, {"name": "machine"}],
        "content": [{"format": "plainText", "text": "mycontent"}],
    }
    assert message == expected

    client.send_message(
        "mycontent", category="comment", tags=["commenttag"], software="mysoft"
    )
    message = messages.get(timeout=10)
    assert message.pop("apikey")
    assert message.pop("creationDate")
    assert message.pop("machine")
    expected = {
        "type": "broadcast",
        "source": "ebs",
        "category": "comment",
        "software": "mysoft",
        "tag": [{"name": "testtag"}, {"name": "machine"}, {"name": "commenttag"}],
        "content": [{"format": "plainText", "text": "mycontent"}],
    }
    assert message == expected

    assert messages.empty()


def test_elogbook_message_beamline_only(elogbook_client):
    client, messages = elogbook_client
    client.send_message("mycontent", category="comment", beamline="id00")
    message = messages.get(timeout=10)
    assert message.pop("apikey")
    assert message.pop("creationDate")
    assert message.pop("machine")
    assert message.pop("software").startswith("pyicat-plus")
    expected = {
        "type": "annotation",
        "category": "comment",
        "content": [{"format": "plainText", "text": "mycontent"}],
        "instrument": "id00",
    }
    assert message == expected
    assert messages.empty()


def test_elogbook_data(elogbook_client):
    client, messages = elogbook_client
    client.send_binary_data(
        b"123", mimetype="application/octet-stream", beamline="id00", proposal="hg123"
    )
    message = messages.get(timeout=10)
    assert message.pop("apikey")
    assert message.pop("creationDate")
    assert message.pop("machine")
    assert message.pop("software").startswith("pyicat-plus")
    data = message.pop("base64")
    data = data.replace("data:application/octet-stream;base64,", "")
    assert base64.b64decode(data.encode()) == b"123"
    expected = {"investigation": "hg123", "instrument": "id00"}
    assert message == expected
    assert messages.empty()


def test_elogbook_data_beamline_only(elogbook_client):
    client, messages = elogbook_client
    client.send_binary_data(
        b"123", mimetype="application/octet-stream", beamline="id00"
    )
    message = messages.get(timeout=10)
    assert message.pop("apikey")
    assert message.pop("creationDate")
    assert message.pop("machine")
    assert message.pop("software").startswith("pyicat-plus")
    data = message.pop("base64")
    data = data.replace("data:application/octet-stream;base64,", "")
    assert base64.b64decode(data.encode()) == b"123"
    expected = {"instrument": "id00"}
    assert message == expected
    assert messages.empty()



from collections.abc import Iterable, Mapping
from numbers import Number
from typing import Union
from datetime import datetime


def serialize_metadata(
    obj: Union[str, bytes, Number, None, Mapping, Iterable], **_recursive_info
) -> Union[str, None, dict]:
    """Serialize metadata for ICAT"""
    _recursive_info.setdefault("depth", 0)
    _recursive_info.setdefault("iterable_depth", 0)
    if isinstance(obj, str):
        return obj
    elif isinstance(obj, bytes):
        return obj.decode()
    elif isinstance(obj, Number):
        return str(obj)
    elif isinstance(obj, datetime):
        return obj.isoformat()
    elif isinstance(obj, Mapping):
        if _recursive_info["depth"]:
            raise TypeError(obj)
        _recursive_info["depth"] += 1
        return {
            serialize_metadata(k, **_recursive_info): serialize_metadata(
                v, **_recursive_info
            )
            for k, v in obj.items()
            if v is not None
        }
    elif isinstance(obj, Iterable):
        if _recursive_info["iterable_depth"] > 1:
            raise ValueError("")
        if all(isinstance(v, (str, bytes, Number)) or v is None for v in obj):
            sep = ","
        else:
            sep = " "
        _recursive_info["depth"] += 1
        _recursive_info["iterable_depth"] += 1
        return sep.join(
            [serialize_metadata(v, **_recursive_info) for v in obj if v is not None]
        )
    elif obj is None:
        return None
    else:
        raise TypeError(obj)



"""Query pools let a query run in the background when it
doesn't return within a given timeout. In that case the
result of the previous query is returned or raised. If
there is no result, the default value is returned.
"""

from .. import gevent

if gevent is None:
    from .threading import QueryPool  # noqa F401
else:
    from .gevent import QueryPool  # noqa F401



from ..utils.url import normalize_url


def test_normlize_url():
    url = "sub.name.root"
    expected = "//sub.name.root"
    assert normalize_url(url) == expected

    url = "sub.name.root"
    expected = "http://sub.name.root"
    assert normalize_url(url, default_scheme="http") == expected
    url = "https://sub.name.root"
    expected = "https://sub.name.root"
    assert normalize_url(url, default_scheme="http") == expected

    url = "sub.name.root"
    expected = "//sub.name.root:80"
    assert normalize_url(url, default_port=80) == expected
    url = "https://sub.name.root"
    expected = "https://sub.name.root:80"
    assert normalize_url(url, default_port=80) == expected

    url = "sub.name.root"
    expected = "http://sub.name.root:80"
    assert normalize_url(url, default_scheme="http", default_port=80) == expected
    url = "https://sub.name.root:8080"
    expected = "https://sub.name.root:8080"
    assert normalize_url(url, default_port=80) == expected

    url = "sub.name.root:8080"
    expected = "http://sub.name.root:8080"
    assert normalize_url(url, default_scheme="http", default_port=80) == expected
    url = "https://sub.name.root"
    expected = "https://sub.name.root:80"
    assert normalize_url(url, default_scheme="http", default_port=80) == expected

    url = "sub.name.root:8080"
    expected = "http://sub.name.root:8080"
    assert normalize_url(url, default_scheme="http", default_port=80) == expected
    url = "https://sub.name.root:8080"
    expected = "https://sub.name.root:8080"
    assert normalize_url(url, default_scheme="http", default_port=80) == expected



try:
    import gevent
except ImportError:
    gevent = None
else:
    from gevent.monkey import is_anything_patched

    if not is_anything_patched():
        gevent = None


if gevent is None:
    import threading
    from subprocess import TimeoutExpired
    from queue import Queue, Empty

    def spawn(func, *args, **kwargs):
        thread = threading.Thread(target=func, args=args, kwargs=kwargs)
        thread.start()
        return thread

    def wait_process(process, timeout) -> bool:
        """
        :param process: A process object from `subprocess` or `psutil`
        """
        try:
            process.wait(timeout)
            return True
        except (TimeoutError, TimeoutExpired):
            return False

else:
    from gevent import spawn  # noqa F401
    from gevent.queue import Queue, Empty

    def wait_process(process, timeout) -> bool:
        """
        :param process: A process object from `subprocess` or `psutil`
        """
        try:
            with gevent.Timeout(timeout) as local_timeout:
                # gevent timeout has to be used here
                # See https://github.com/gevent/gevent/issues/622
                process.wait()
            return True
        except gevent.Timeout as raised_timeout:
            if local_timeout is not raised_timeout:
                raise
            return False


def flush_queue(q: Queue):
    while True:
        try:
            yield q.get(timeout=0)
        except Empty:
            break



import sys
import argparse
from . import store_raw
from ..client.main import IcatClient


def main(argv=None):
    if argv is None:
        argv = sys.argv

    parser = argparse.ArgumentParser(description="Register raw data with ICAT")

    store_raw.add_store_parameters(parser)
    add_process_parameters(parser)

    args = parser.parse_args(argv[1:])

    store_raw.apply_store_parameters(args)
    apply_process_parameters(args)

    client = IcatClient(metadata_urls=args.metadata_urls)
    client.store_processed_data(
        beamline=args.beamline,
        proposal=args.proposal,
        dataset=args.dataset,
        path=args.path,
        metadata=args.metadata,
        raw=args.raw,
    )


def add_process_parameters(parser):
    parser.add_argument(
        "--raw",
        action="append",
        required=True,
        help="Raw dataset directories",
    )


def apply_process_parameters(args):
    pass


if __name__ == "__main__":
    sys.exit(main())



import sys
import socket
import subprocess
from contextlib import contextmanager

import pytest

from . import proc
from . import tcp
from ...concurrency import flush_queue
from ...client.archive import IcatArchiveStatusClient
from ...client.elogbook import IcatElogbookClient
from ...client.metadata import IcatMetadataClient


__all__ = [
    # Backends
    "icat_data_dir",
    "cleanup_backend",
    "stomp_server",
    "icatplus_server",
    "activemq_rest_server",
    "icat_backend",
    # Read from the backend
    "icat_subscriber",
    "icat_subscriber_archive",
    "session_icat_logbook_subscriber",
    "icat_logbook_subscriber",
    # Write to the backend
    "session_icat_publisher",
    "icat_publisher",
    "session_elogbook_client",
    "elogbook_client",
    "session_elogbook_ebs_client",
    "elogbook_ebs_client",
    "icat_metadata_client",
    "icat_archive_client",
]


@pytest.fixture(scope="session")
def icat_data_dir(tmpdir_factory):
    """Directory where the ICAT dataset backend icat_subscriber can store state."""
    yield tmpdir_factory.mktemp("datasets")


@pytest.fixture()
def cleanup_backend(icat_data_dir):
    for dataset in icat_data_dir.listdir():
        dataset.remove()


@pytest.fixture(scope="session")
def stomp_server():
    """One of the ICAT backends (for dataset metadata)"""
    port = tcp.get_open_port()
    # Add arguments ["--debug", "TEXT"] for debugging
    p = subprocess.Popen(["coilmq", "-b", "0.0.0.0", "-p", str(port)])
    tcp.wait_tcp_online("localhost", port)
    try:
        yield "localhost", port
    finally:
        proc.wait_terminate(p)


@pytest.fixture(scope="session")
def activemq_rest_server():
    """One of the ICAT backends (for queue monitoring)"""
    port = tcp.get_open_port()
    p = subprocess.Popen(
        [
            sys.executable,
            "-u",
            "-m",
            "pyicat_plus.tests.servers.activemq_rest_server",
            f"--port={port}",
        ]
    )
    tcp.wait_tcp_online("localhost", port)
    try:
        yield "localhost", port
    finally:
        proc.wait_terminate(p)


@pytest.fixture(scope="session")
def icat_backend(stomp_server, activemq_rest_server, icatplus_server):
    """Starts all icat backends"""
    host, port = stomp_server
    metadata_urls = [f"{host}:{port}"]
    port, _ = icatplus_server
    elogbook_url = f"http://localhost:{port}"
    host, port = activemq_rest_server
    metadata_queue_monitor_port = port
    icat_servers = {
        "metadata_urls": metadata_urls,
        "elogbook_url": elogbook_url,
        "metadata_queue_monitor_port": metadata_queue_monitor_port,
        "elogbook_timeout": 5,
        "feedback_timeout": 5,
        "queue_timeout": 5,
    }
    yield icat_servers


@pytest.fixture(scope="session")
def icat_subscriber(stomp_server, icat_data_dir):
    with _icat_subscriber(stomp_server, icat_data_dir, "/queue/icatIngest") as messages:
        yield messages


@pytest.fixture(scope="session")
def icat_subscriber_archive(stomp_server, icat_data_dir):
    with _icat_subscriber(
        stomp_server, icat_data_dir, "/queue/icatArchiveRestoreStatus"
    ) as messages:
        yield messages


@contextmanager
def _icat_subscriber(stomp_server, icat_data_dir, queue):
    """Receive messages from the stomp_server backend"""
    with tcp.tcp_message_server() as (port_out, messages):
        host, port = stomp_server
        p = subprocess.Popen(
            [
                sys.executable,
                "-u",
                "-m",
                "pyicat_plus.tests.servers.stomp_subscriber",
                f"--host={host}",
                f"--port={port}",
                f"--port_out={port_out}",
                f"--queue={queue}",
                f"--icat_data_dir={icat_data_dir}",
            ]
        )
        try:
            assert messages.get(timeout=5) == "LISTENING"
            yield messages
        finally:
            proc.wait_terminate(p)


@pytest.fixture(scope="session")
def icatplus_server(icat_subscriber, icat_data_dir):
    """ICAT backend for the e-logbook and dataset queries.
    The icat_subscriber is needed to save published datasets
    in files so that the icatplus_server can use it when
    asked for the all published datasets.
    """
    with tcp.tcp_message_server("json") as (port_out, messages):
        port = tcp.get_open_port()
        p = subprocess.Popen(
            [
                sys.executable,
                "-u",
                "-m",
                "pyicat_plus.tests.servers.icatplus_server",
                f"--port={port}",
                f"--port_out={port_out}",
                f"--icat_data_dir={icat_data_dir}",
            ]
        )
        tcp.wait_tcp_online("localhost", port)
        try:
            assert messages.get(timeout=5) == {"STATUS": "LISTENING"}
            yield port, messages
        finally:
            proc.wait_terminate(p)


@pytest.fixture(scope="session")
def session_icat_logbook_subscriber(icatplus_server):
    """Receive messages from the icatplus_server backend"""
    _, messages = icatplus_server
    return messages


@pytest.fixture()
def icat_logbook_subscriber(session_icat_logbook_subscriber, cleanup_backend):
    """Receive messages from the icatplus_server backend"""
    with assert_no_messages(session_icat_logbook_subscriber):
        yield session_icat_logbook_subscriber


@contextmanager
def assert_no_messages(messages):
    for _ in flush_queue(messages):
        pass
    try:
        yield
    finally:
        try:
            assert messages.empty(), "not all messages have been validated"
        finally:
            for msg in flush_queue(messages):
                print(f"\nUnvalidated message: {msg}")


@pytest.fixture(scope="session")
def session_icat_publisher(stomp_server):
    """Sends messages to the /queue/icatIngest queue of stomp_server"""
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    port_in = tcp.get_open_port()
    sock.bind(("localhost", port_in))
    sock.listen()

    # Redirect messages from socket to STOMP server
    host, port = stomp_server
    p = subprocess.Popen(
        [
            sys.executable,
            "-u",
            "-m",
            "pyicat_plus.tests.servers.stomp_publisher",
            f"--host={host}",
            f"--port={port}",
            f"--port_in={port_in}",
            "--queue=/queue/icatIngest",
        ]
    )
    try:
        conn, addr = sock.accept()
        with conn:
            yield conn
    finally:
        sock.close()
        proc.wait_terminate(p)


@pytest.fixture()
def icat_publisher(session_icat_publisher, cleanup_backend):
    """Sends messages to the /queue/icatIngest queue of stomp_server"""
    yield session_icat_publisher


@pytest.fixture(scope="session")
def session_elogbook_client(icatplus_server):
    """Client to publish e-logbook messages and a queue that
    receives everything the ICAT backend receives.
    """
    port, messages = icatplus_server
    client = IcatElogbookClient(f"http://localhost:{port}")
    yield client, messages


@pytest.fixture()
def elogbook_client(session_elogbook_client, cleanup_backend):
    """Client to publish e-logbook messages and a queue that
    receives everything the ICAT backend receives.
    """
    client, messages = session_elogbook_client
    with assert_no_messages(messages):
        yield client, messages


@pytest.fixture(scope="session")
def session_elogbook_ebs_client(icatplus_server):
    """Client to publish e-logbook messages and a queue that
    receives everything the ICAT backend receives.
    """
    port, messages = icatplus_server
    client = IcatElogbookClient(
        f"http://localhost:{port}",
        tags=["testtag", "machine"],
        software="testsoft",
        source="ebs",
        type="broadcast",
    )
    yield client, messages


@pytest.fixture()
def elogbook_ebs_client(session_elogbook_ebs_client, cleanup_backend):
    """Client to publish e-logbook messages and a queue that
    receives everything the ICAT backend receives.
    """
    client, messages = session_elogbook_ebs_client
    with assert_no_messages(messages):
        yield client, messages


@pytest.fixture()
def icat_metadata_client(
    stomp_server, activemq_rest_server, icat_subscriber, cleanup_backend
):
    """Client to publish ICAT metadata and a queue that
    receives everything the ICAT backend receives.
    """
    _, jport = activemq_rest_server
    host, port = stomp_server
    messages = icat_subscriber
    client = IcatMetadataClient([f"{host}:{port}"], monitor_port=jport)
    yield client, messages


@pytest.fixture()
def icat_archive_client(
    stomp_server, activemq_rest_server, icat_subscriber_archive, cleanup_backend
):
    """Client to update dataset archive/restore status and a queue that
    receives everything the ICAT backend receives.
    """
    _, jport = activemq_rest_server
    host, port = stomp_server
    messages = icat_subscriber_archive
    client = IcatArchiveStatusClient([f"{host}:{port}"], monitor_port=jport)
    yield client, messages



from datetime import datetime
import xml.etree.ElementTree as etree
from .serialize import serialize_metadata

ICAT_NAMESPACE_URL = "http://www.esrf.fr/icat"

etree.register_namespace("tns", ICAT_NAMESPACE_URL)


def root_node(name: str, **kw):
    return etree.Element(f"{{{ICAT_NAMESPACE_URL}}}{name}", **kw)


def child_node(parent, name: str, **kw):
    return etree.SubElement(parent, f"{{{ICAT_NAMESPACE_URL}}}{name}", **kw)


def encode_node_data(data) -> str:
    text = serialize_metadata(data)
    if not isinstance(text, str):
        raise ValueError(data)
    return text


def data_node(parent, name: str, data, **kw):
    node = child_node(parent, name, **kw)
    node.text = encode_node_data(data)


def parameter_node(parent, name: str, value, **kw):
    node = child_node(parent, "parameter", **kw)
    data_node(node, "name", name)
    data_node(node, "value", value)


def dataset_as_xml(
    beamline: str, proposal: str, dataset: str, path: str, metadata: dict = None
):
    assert beamline, "ICAT requires the beamline name"
    assert proposal, "ICAT requires the proposal name"
    assert dataset, "ICAT requires the dataset name"
    assert path, "ICAT requires the dataset path"
    if metadata is None:
        metadata = dict()

    # Required metadata
    assert "Sample_name" in metadata, "ICAT metadata field 'Sample_name' is missing"

    # Metadata with defaults
    if "startDate" not in metadata:
        metadata["startDate"] = datetime.now()
    if "endDate" not in metadata:
        metadata["endDate"] = datetime.now()

    root = root_node("dataset", attrib={"complete": "true"})
    data_node(root, "investigation", proposal)
    data_node(root, "instrument", beamline)
    data_node(root, "name", dataset)
    data_node(root, "location", path)

    metadata = serialize_metadata(metadata)
    for name, value in metadata.items():
        parameter_node(root, name, value)
        # Metadata included in the XML tree
        if name == "Sample_name":
            sample = child_node(root, "sample")
            data_node(sample, "name", value)
        elif name == "startDate":
            data_node(root, "startDate", value)
        elif name == "endDate":
            data_node(root, "endDate", value)

    return root


def investigation_as_xml(beamline: str, proposal: str, start_datetime=None):
    root = root_node("investigation")
    data_node(root, "experiment", proposal)
    data_node(root, "instrument", beamline)
    if start_datetime is None:
        start_datetime = datetime.now()
    data_node(root, "startDate", start_datetime)
    return root



from typing import Any, Callable, Dict, Optional
from time import time
from contextlib import contextmanager
from concurrent.futures import ThreadPoolExecutor, Future, TimeoutError
from threading import RLock

from ...utils.maxsizedict import MaxSizeDict


class QueryPool:
    """Query pool for gevent non-cooperative calls."""

    def __init__(
        self, timeout: Optional[float] = None, maxqueries: Optional[int] = None
    ):
        """
        :param timeout: The default timeout of a call before returning/raising the previous result.
        :param maxqueries: The maximal number of different queries to store results from.
                           A query can differ in terms of function and/or arguments.
        """
        self.__pool = ThreadPoolExecutor(max_workers=10000)
        if timeout is None:
            timeout = 0.1
        self.timeout = timeout
        self.__futures: Dict[tuple, Future] = dict()
        if maxqueries:
            self.__lock = RLock()
            self.__results = MaxSizeDict(maxsize=maxqueries)
        else:
            self.__lock = None
            self.__results = dict()

    @contextmanager
    def _lock(self):
        if self.__lock is None:
            yield
        else:
            with self.__lock:
                yield

    def execute(
        self,
        query: Callable,
        args: Optional[tuple] = tuple(),
        kwargs: Optional[dict] = None,
        timeout: Optional[float] = None,
        default=None,
    ) -> Any:
        """
        :param query:
        :param args: positional arguments
        :param kwargs: named arguments
        :param timeout: the timeout of a call before returning/raising the previous result
        :param default: the default value in case there is no previous result
        :returns: the result of the query or the default value
        :raises: the exception from the query
        """
        if kwargs is None:
            kwargs = dict()
        call_id = query, args, tuple(kwargs.items())
        future = self.__futures.get(call_id)
        if future is None:

            def wrapper():
                try:
                    result = query(*args, **kwargs)
                    with self._lock():
                        self.__results[call_id] = False, result
                except BaseException as e:
                    with self._lock():
                        self.__results[call_id] = True, e
                    raise
                finally:
                    with self._lock():
                        self.__futures.pop(call_id, None)

            with self._lock():
                self.__futures[call_id] = future = self.__pool.submit(wrapper)

        if timeout is None:
            timeout = self.timeout
        try:
            future.result(timeout=timeout)
        except TimeoutError:
            pass
        result = self.__results.get(call_id, None)
        if result is None:
            return default
        is_error, result = result
        if is_error:
            raise result
        return result

    def wait(self, timeout=None) -> bool:
        """
        :param timeout:
        :returns: `True` when all queries finished, `False` otherwise
        """
        try:
            with self._lock():
                futures = list(self.__futures.values())
            for future in futures:
                t0 = time()
                future.result(timeout=timeout)
                timeout = max(timeout - time() + t0, 0)
        except TimeoutError:
            return False
        return True



import os
import sys
import argparse
from glob import glob
from ..client.main import IcatClient


def main(argv=None):
    if argv is None:
        argv = sys.argv

    parser = argparse.ArgumentParser(description="Register stored data with ICAT")
    add_store_parameters(parser)
    args = parser.parse_args(argv[1:])
    apply_store_parameters(args)

    client = IcatClient(metadata_urls=args.metadata_urls)

    for filename in args.files:
        print("Register", filename)
        client.store_dataset_from_file(filename)


def add_store_parameters(parser):
    parser.add_argument("filter", help="File search filter")

    parser.add_argument(
        "--queue",
        dest="metadata_urls",
        action="append",
        help="ActiveMQ queue URLS. Example: bcu-mq-01.esrf.fr:61613",
        default=[],
    )


def apply_store_parameters(args):
    args.files = sorted(glob(args.filter), key=os.path.getmtime)

    if not args.metadata_urls:
        args.metadata_urls = ["bcu-mq-01.esrf.fr:61613", "bcu-mq-02.esrf.fr:61613"]


if __name__ == "__main__":
    sys.exit(main())



import pytest
from ..concurrency import gevent
from ..concurrency.query_pool import QueryPool

if gevent is None:
    gevent = None
    from time import sleep as noncooperative_sleep

    cooperative_sleep = noncooperative_sleep
else:
    cooperative_sleep = gevent.sleep
    noncooperative_sleep = gevent.monkey.get_original("time", "sleep")


def query(seconds, cooperative=True):
    print(f"sleep for {seconds} seconds ...")
    try:
        if cooperative:
            cooperative_sleep(seconds)
        else:
            noncooperative_sleep(seconds)
    except BaseException as e:
        print(f"sleep interrupted: {e}")
        raise
    print("sleep done")


@pytest.mark.timeout(5)
def test_query_success():
    """test executing a query which returns a value"""
    pool = QueryPool(timeout=0.1)

    def myfunc(a, b=0):
        query(a + b)
        return True

    while not pool.execute(myfunc, args=(0.5,), kwargs={"b": 0.5}):
        pass  # every iteration waits for 0.1 seconds


@pytest.mark.timeout(5)
def test_query_exception():
    """test executing a query which raises an exception"""
    pool = QueryPool(timeout=0.1)

    def myfunc(a, b=0):
        query(a + b)
        raise RuntimeError

    with pytest.raises(RuntimeError):
        while not pool.execute(myfunc, args=(0.5,), kwargs={"b": 0.5}):
            pass  # every iteration waits for 0.1 seconds


def test_query_wait():
    """test executing a query which returns a value"""
    pool = QueryPool(timeout=0.1)

    finished = False

    def myfunc():
        nonlocal finished
        query(1)
        finished = True

    pool.execute(myfunc)

    assert pool.wait(timeout=3), "query did not finish in time"

    assert finished, "query did not finish"


@pytest.mark.skipif(gevent is None, reason="requires gevent for now")
def test_query_cancel():
    """test canceling a query"""
    pool = QueryPool(timeout=0.1)

    raised = False

    def myfunc():
        nonlocal raised
        try:
            query(2)
        except gevent.GreenletExit:
            raised = True
            raise

    pool.execute(myfunc)

    assert pool.cancel(timeout=3), "could not cancel the query in time"

    assert raised, "query was not cancelled"


@pytest.mark.timeout(5)
def test_query_maxqueries1():
    """test executing a query which returns a value"""
    pool = QueryPool(timeout=0.1, maxqueries=1)

    def myfunc(a, b=0):
        query(a + b)
        return True

    while not pool.execute(myfunc, args=(0.1,), kwargs={"b": 0.1}):
        pass  # every iteration waits for 0.1 seconds
    while not pool.execute(myfunc, args=(0.1,), kwargs={"b": 0.2}):
        pass  # every iteration waits for 0.1 seconds

    previous = pool.execute(myfunc, timeout=0, args=(0.1,), kwargs={"b": 0.2})
    assert previous, "pool size not large enough"
    missing = pool.execute(myfunc, timeout=0, args=(0.1,), kwargs={"b": 0.1})
    assert missing is None, "pool size too large"
    assert pool.wait(timeout=3)


@pytest.mark.timeout(5)
def test_query_maxqueries2():
    """test executing a query which returns a value"""
    pool = QueryPool(timeout=0.1, maxqueries=3)

    def myfunc(a, b=0):
        query(a + b)
        return True

    while not pool.execute(myfunc, args=(0.1,), kwargs={"b": 0.1}):
        pass  # every iteration waits for 0.1 seconds
    while not pool.execute(myfunc, args=(0.1,), kwargs={"b": 0.2}):
        pass  # every iteration waits for 0.1 seconds

    previous = pool.execute(myfunc, timeout=0, args=(0.1,), kwargs={"b": 0.2})
    assert previous, "pool size not large enough"
    previous = pool.execute(myfunc, timeout=0, args=(0.1,), kwargs={"b": 0.1})
    assert previous, "pool size not large enough"
    assert pool.wait(timeout=3)


@pytest.mark.skipif(gevent is None, reason="requires gevent for now")
def test_noncooperative_query_in_cooperative_pool():
    """test executing a non-cooperative query in a cooperative pool (improper usage)"""
    pool = QueryPool(timeout=0.1)

    call_count = 0

    def myfunc(**kw):
        nonlocal call_count
        call_count += 1
        if call_count > 1:
            # Note: this exception will also be printed when the test passed
            raise RuntimeError("the gevent loop was blocked")
        query(**kw)

    with pytest.raises(RuntimeError, match="the gevent loop was blocked"):
        with gevent.Timeout(0.5, "the gevent loop was not blocked"):
            while not pool.execute(myfunc, kwargs={"seconds": 2, "cooperative": False}):
                pass  # every iteration waits for 0.1 seconds

    assert pool.cancel(timeout=4)



from urllib.parse import urlparse


def normalize_url(
    url: str,
    default_scheme: str = None,
    default_port: int = None,
    absolute: bool = True,
) -> str:
    if not url:
        raise ValueError("URL is missing")
    if absolute and "//" not in url:
        url = "//" + url
    result = urlparse(url)
    scheme, netloc, *others = result
    if not netloc:
        raise ValueError(url, "URL is missing a network location")
    if not scheme and default_scheme:
        scheme = default_scheme
    if default_port and not result.port:
        netloc = f"{result.hostname}:{default_port}"
    newurl = type(result)(scheme, netloc, *others)
    return newurl.geturl()



import sys
import time
from datetime import datetime, timedelta
import socketserver
import logging

from pyicat_plus.tests.servers.utils import basic_config
from pyicat_plus.tests.servers.utils import ReuseAddrTCPServer


logger = logging.getLogger("ACTIVEMQ REST SERVER")
basic_config(
    logger=logger,
    level=logging.DEBUG,
    format="%(asctime)s - %(name)s - %(levelname)s - %(message)s",
)
logger.addHandler(logging.StreamHandler(sys.stdout))


class MyTCPRequestHandler(socketserver.StreamRequestHandler):
    def handle(self):
        request = self._read_request()
        queues = ["icatIngest", "icatArchiveRestoreStatus"]
        expected_requests = list(
            map(
                lambda q: f"GET /api/jolokia/read/org.apache.activemq:type=Broker,brokerName=metadata,destinationType=Queue,destinationName={q}/ConsumerCount".encode(
                    "utf-8"
                ),
                queues,
            )
        )
        if len([e for e in expected_requests if e in request]) > 0:
            logger.info(f"Send response to {self.client_address[0]}")
            self._send_response()
        elif request:
            logger.info(f"Unknown request\n {request}")
            raise RuntimeError("Unknown request")

    def _read_request(self):
        buff = bytearray(16384)
        request = b""
        try:
            n = self.rfile.readinto1(buff)
            request = bytes(buff[0:n])
        except Exception as e:
            raise RuntimeError("Error reading request") from e
        return request

    def _send_response(self):
        now = datetime.now()
        t1 = now + timedelta(hours=5)
        out = (
            b"HTTP/1.1 200 OK\r\nContent-Type: text/plain;charset=UTF-8\r\nCache-Control: no-cache\r\nPragma: no-cache\r\nDate: "
            + now.strftime("%a, %d %b %Y %H:%M:%S GTM").encode()
            + b"\r\nExpires: "
            + t1.strftime("%a, %d %b %Y %H:%M:%S GTM").encode()
            + b'\r\nConnection: close\r\nServer: Jetty(7.6.9.v20130131)\r\n\r\n{"timestamp":'
            + str(int(time.time())).encode()
            + b',"status":200,"request":{"mbean":"org.apache.activemq:brokerName=metadata,destinationName=icatIngest,destinationType=Queue,type=Broker","attribute":"ConsumerCount","type":"read"},"value":6}'
        )
        self.wfile.write(out)


def main(port=8778):
    # Create a TCP Server instance
    aServer = ReuseAddrTCPServer(("localhost", port), MyTCPRequestHandler)

    # Listen forever
    logger.info("CTRL-C to stop")
    try:
        aServer.serve_forever()
    finally:
        logger.info("Exit.")


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser(description="ActiveMQ REST server")
    parser.add_argument("--port", default=8778, type=int, help="server port")
    args = parser.parse_args()
    main(port=args.port)



from ..metadata.nexus import create_nxtreedict
from .utils.compare import deep_compare


def test_icat_metadata_to_nexus(icat_fields):
    metadict = dict()

    def getter(key):
        nonlocal metadict
        return metadict[key]

    def setter(key, value):
        nonlocal metadict
        metadict[key] = value

    metadata = icat_fields.namespace(getter=getter, setter=setter)

    metadata.instrument.detector01.name = "diode1"
    metadata.instrument.detector02.name = "diode2"

    metadata.instrument.variables.name = ["roby", "robz"]
    metadata.instrument.variables.value = [0, 0]

    metadata.instrument.insertion_device.gap.name = ["roby", "robz"]
    metadata.instrument.insertion_device.gap.value = [0, 0]

    metadata.instrument.primary_slit.name = "primary_slit"
    metadata.instrument.primary_slit.horizontal_gap = 0
    metadata.instrument.primary_slit.horizontal_offset = 0
    metadata.instrument.primary_slit.vertical_gap = 0
    metadata.instrument.primary_slit.vertical_offset = 0

    metadata.sample.name = "sample"
    metadata.sample.positioners.name = "roby"
    metadata.sample.positioners.value = 0

    metadata.FLUO.i0 = 1
    metadata.FLUO.measurement.i0_start = 0.5
    metadata.definition = "FLUO"

    nxtreedict = create_nxtreedict(metadict, icat_fields)
    expected = {
        "@NX_class": "NXentry",
        "FLUO": {
            "@NX_class": "NXsubentry",
            "i0": 1.0,
            "measurement": {
                "@NX_class": "NXcollection",
                "i0_start": 0.5,
                "i0_start@units": "photons/s",
            },
        },
        "instrument": {
            "@NX_class": "NXinstrument",
            "variables": {
                "@NX_class": "NXcollection",
                "name": ["roby", "robz"],
                "value": [0, 0],
            },
            "insertion_device": {
                "@NX_class": "NXinsertion_device",
                "gap": {
                    "@NX_class": "NXpositioner",
                    "name": ["roby", "robz"],
                    "value": [0, 0],
                },
            },
            "primary_slit": {
                "@NX_class": "NXslit",
                "name": "primary_slit",
                "horizontal_gap": 0,
                "horizontal_offset": 0,
                "vertical_gap": 0,
                "vertical_offset": 0,
            },
            "detector01": {"@NX_class": "NXdetector", "name": "diode1"},
            "detector02": {"@NX_class": "NXdetector", "name": "diode2"},
        },
        "sample": {
            "@NX_class": "NXsample",
            "name": "sample",
            "positioners": {"@NX_class": "NXpositioner", "name": "roby", "value": 0},
        },
        "definition": "FLUO",
    }
    deep_compare(nxtreedict, expected)



from typing import List, Mapping, Optional, Sequence, Tuple

import numpy

from .archive import IcatArchiveStatusClient, StatusLevel, StatusType
from .elogbook import IcatElogbookClient
from .interface import DatasetId, IcatClientInterface
from .investigation import IcatInvestigationClient
from .metadata import IcatMetadataClient


class IcatClient(IcatClientInterface):
    """Direct communication with ICAT: e-logbook, metadata and archive status"""

    def __init__(
        self,
        metadata_urls: Optional[list] = None,
        elogbook_url: Optional[str] = None,
        elogbook_token: Optional[str] = "elogbook-00000000-0000-0000-0000-000000000000",
        metadata_queue: Optional[str] = "icatIngest",
        metadata_queue_monitor_port: Optional[int] = None,
        elogbook_timeout: Optional[float] = None,
        feedback_timeout: Optional[float] = None,
        queue_timeout: Optional[float] = None,
        beamline: Optional[str] = None,
        proposal: Optional[str] = None,
        elogbook_metadata: Optional[Mapping] = None,
        archive_urls: Optional[list] = None,
        archive_queue: Optional[str] = "icatArchiveRestoreStatus",
        archive_queue_monitor_port: Optional[int] = None,
    ):
        self.current_proposal = proposal
        self.current_beamline = beamline
        self.current_dataset = None
        self.current_path = None
        self.current_dataset_metadata = None
        if metadata_urls:
            self._metadata_client = IcatMetadataClient(
                queue_urls=metadata_urls,
                queue_name=metadata_queue,
                monitor_port=metadata_queue_monitor_port,
                timeout=queue_timeout,
            )
        else:
            self._metadata_client = None
        if archive_urls:
            self._archive_client = IcatArchiveStatusClient(
                queue_urls=archive_urls,
                queue_name=archive_queue,
                monitor_port=archive_queue_monitor_port,
                timeout=queue_timeout,
            )
        else:
            self._archive_client = None
        if elogbook_url and elogbook_token:
            self._investigation_client = IcatInvestigationClient(
                url=elogbook_url, api_key=elogbook_token, timeout=feedback_timeout
            )
            self._elogbook_client = IcatElogbookClient(
                url=elogbook_url,
                api_key=elogbook_token,
                timeout=elogbook_timeout,
                payload=elogbook_metadata,
            )
        else:
            self._investigation_client = None
            self._elogbook_client = None

    @property
    def metadata_client(self):
        if self._metadata_client is None:
            raise RuntimeError("The message queue URL's are missing")
        return self._metadata_client

    @property
    def investigation_client(self):
        if self._investigation_client is None:
            raise RuntimeError("The ICAT+ URL and/or token are missing")
        return self._investigation_client

    @property
    def elogbook_client(self):
        if self._elogbook_client is None:
            raise RuntimeError("The ICAT+ URL and/or token are missing")
        return self._elogbook_client

    @property
    def current_proposal(self):
        return self.__current_proposal

    @current_proposal.setter
    def current_proposal(self, value: Optional[str]):
        self.__current_proposal = value

    @property
    def current_beamline(self):
        return self.__current_beamline

    @current_beamline.setter
    def current_beamline(self, value: Optional[str]):
        self.__current_beamline = value

    @property
    def current_dataset(self):
        return self.__current_dataset

    @current_dataset.setter
    def current_dataset(self, value: Optional[str]):
        self.__current_dataset = value

    @property
    def current_dataset_metadata(self):
        return self.__current_dataset_metadata

    @current_dataset_metadata.setter
    def current_dataset_metadata(self, value: Optional[dict]):
        self.__current_dataset_metadata = value

    @property
    def current_path(self):
        return self.__current_path

    @current_path.setter
    def current_path(self, value: Optional[str]):
        self.__current_path = value

    @property
    def archive_client(self):
        if self._archive_client is None:
            raise RuntimeError("The message queue URL's are missing")
        return self._archive_client

    def send_message(
        self,
        msg: str,
        msg_type="comment",
        beamline: Optional[str] = None,
        proposal: Optional[str] = None,
        dataset: Optional[str] = None,
        beamline_only: Optional[bool] = None,
        **payload,
    ):
        if beamline_only:
            proposal = None
        elif proposal is None:
            proposal = self.current_proposal
        if beamline is None:
            beamline = self.current_beamline
        if beamline_only:
            dataset = None
        elif dataset is None:
            dataset = self.current_dataset
        self.elogbook_client.send_message(
            message=msg,
            category=msg_type,
            beamline=beamline,
            proposal=proposal,
            dataset=dataset,
            **payload,
        )

    def send_binary_data(
        self,
        data: bytes,
        mimetype: Optional[str] = None,
        beamline: Optional[str] = None,
        proposal: Optional[str] = None,
        beamline_only: Optional[bool] = None,
        **payload,
    ):
        if beamline_only:
            proposal = None
        elif proposal is None:
            proposal = self.current_proposal
        if beamline is None:
            beamline = self.current_beamline
        self.elogbook_client.send_binary_data(
            data, mimetype=mimetype, beamline=beamline, proposal=proposal, **payload
        )

    def send_text_file(
        self,
        filename: str,
        beamline: Optional[str] = None,
        proposal: Optional[str] = None,
        dataset: Optional[str] = None,
        beamline_only: Optional[bool] = None,
        **payload,
    ):
        if beamline_only:
            proposal = None
        elif proposal is None:
            proposal = self.current_proposal
        if beamline is None:
            beamline = self.current_beamline
        if beamline_only:
            dataset = None
        elif dataset is None:
            dataset = self.current_dataset
        self.elogbook_client.send_text_file(
            filename, beamline=beamline, proposal=proposal, dataset=dataset, **payload
        )

    def send_binary_file(
        self,
        filename: str,
        beamline: Optional[str] = None,
        proposal: Optional[str] = None,
        beamline_only: Optional[bool] = None,
        **payload,
    ):
        if beamline_only:
            proposal = None
        elif proposal is None:
            proposal = self.current_proposal
        if beamline is None:
            beamline = self.current_beamline
        self.elogbook_client.send_binary_file(
            filename, beamline=beamline, proposal=proposal, **payload
        )

    def start_investigation(
        self,
        beamline: Optional[str] = None,
        proposal: Optional[str] = None,
        start_datetime=None,
    ):
        if proposal is None:
            proposal = self.current_proposal
        else:
            self.current_proposal = proposal
        if beamline is None:
            beamline = self.current_beamline
        else:
            self.current_beamline = beamline
        self.metadata_client.start_investigation(
            beamline=beamline, proposal=proposal, start_datetime=start_datetime
        )

    def store_dataset(
        self,
        beamline: Optional[str] = None,
        proposal: Optional[str] = None,
        dataset: Optional[str] = None,
        path: Optional[str] = None,
        metadata: dict = None,
        store_filename: Optional[str] = None,
    ):
        if proposal is None:
            proposal = self.current_proposal
        if beamline is None:
            beamline = self.current_beamline
        if dataset is None:
            dataset = self.current_dataset
        if path is None:
            path = self.current_path
        if metadata is None:
            metadata = self.current_dataset_metadata
            if metadata is None:
                metadata = dict()
        if store_filename:
            self.metadata_client.store_metadata(
                store_filename,
                beamline=beamline,
                proposal=proposal,
                dataset=dataset,
                path=path,
                metadata=metadata,
            )
        else:
            self.metadata_client.send_metadata(
                beamline=beamline,
                proposal=proposal,
                dataset=dataset,
                path=path,
                metadata=metadata,
            )

    def store_processed_data(
        self,
        beamline: Optional[str] = None,
        proposal: Optional[str] = None,
        dataset: Optional[str] = None,
        path: Optional[str] = None,
        metadata: dict = None,
        raw: Sequence = tuple(),
        store_filename: Optional[str] = None,
    ):
        """The 'raw' argument is shorthand for `metadata = {'input_datasets': ...}`."""
        if metadata is None:
            metadata = self.current_dataset_metadata
            if metadata is None:
                metadata = dict()
        if raw:
            if isinstance(raw, Sequence):
                metadata["input_datasets"] = list(raw)
            else:
                metadata["input_datasets"] = [raw]
        if not metadata.get("input_datasets"):
            raise ValueError("Provide 'raw' dataset directories")
        self.store_dataset(
            beamline=beamline,
            proposal=proposal,
            dataset=dataset,
            path=path,
            metadata=metadata,
            store_filename=store_filename,
        )

    def store_dataset_from_file(self, store_filename: Optional[str] = None):
        self.metadata_client.send_metadata_from_file(store_filename)

    def investigation_info(
        self, beamline: str, proposal: str, timeout: Optional[float] = None
    ) -> Optional[dict]:
        return self.investigation_client.investigation_info(
            beamline=beamline, proposal=proposal, timeout=timeout
        )

    def registered_dataset_ids(
        self, beamline: str, proposal: str, timeout: Optional[float] = None
    ) -> Optional[List[DatasetId]]:
        return self.investigation_client.registered_dataset_ids(
            beamline=beamline, proposal=proposal, timeout=timeout
        )

    def investigation_info_string(
        self, beamline: str, proposal: str, timeout: Optional[float] = None
    ) -> str:
        info = self.investigation_info(
            beamline=beamline, proposal=proposal, timeout=timeout
        )
        if info:
            rows = [(str(k), str(v)) for k, v in info.items()]
            lengths = numpy.array([[len(s) for s in row] for row in rows])
            fmt = "   ".join(["{{:<{}}}".format(n) for n in lengths.max(axis=0)])
            infostr = "ICAT proposal time slot:\n "
            infostr += "\n ".join([fmt.format(*row) for row in rows])
        elif info is None:
            infostr = f"Proposal information currently not available ({self.reason_for_missing_information})"
        else:
            infostr = "Proposal NOT available in the data portal"
        return infostr

    def investigation_summary(
        self, beamline: str, proposal: str, timeout: Optional[float] = None
    ) -> List[Tuple]:
        info = self.investigation_info(
            beamline=beamline, proposal=proposal, timeout=timeout
        )
        keys = ["e-logbook", "data portal"]
        if info:
            rows = [(key, info[key]) for key in keys]
        elif info is None:
            rows = [
                (
                    key,
                    f"Proposal information currently not available ({self.reason_for_missing_information})",
                )
                for key in keys
            ]
        else:
            rows = [(key, "Proposal NOT available in the data portal") for key in keys]
        return rows

    def update_archive_restore_status(
        self,
        dataset_id: int = None,
        type: StatusType = None,
        level: StatusLevel = StatusLevel.INFO,
        message: Optional[str] = None,
    ):
        self.archive_client.send_archive_status(
            dataset_id=dataset_id, type=type, level=level, message=message
        )

    @property
    def expire_datasets_on_close(self) -> bool:
        return False

    @property
    def reason_for_missing_information(self) -> str:
        return "ICAT communication timeout"



from enum import Enum
from datetime import datetime
import requests
import base64
import mimetypes
import socket
from urllib.parse import urljoin
from typing import Optional, Iterable, List
import logging

from ..utils.url import normalize_url
from .. import __version__

logger = logging.getLogger(__name__)

MessageCategory = Enum("MessageCategory", "debug info error commandLine comment")
MessageType = Enum("MessageType", "annotation notification")

MessageCategoryMapping = {
    "debug": MessageCategory.debug,
    "info": MessageCategory.info,
    "warning": MessageCategory.error,
    "warn": MessageCategory.error,
    "error": MessageCategory.error,
    "critical": MessageCategory.error,
    "fatal": MessageCategory.error,
    "command": MessageCategory.commandLine,
    "comment": MessageCategory.comment,
}

MessageTypeMapping = {
    MessageCategory.debug: MessageType.notification,
    MessageCategory.info: MessageType.notification,
    MessageCategory.error: MessageType.notification,
    MessageCategory.commandLine: MessageType.notification,
    MessageCategory.comment: MessageType.annotation,
}


class IcatElogbookClient:
    """Client for the e-logbook part of the ICAT+ REST API.

    REST API docs:
    https://icatplus.esrf.fr/api-docs/

    The ICAT+ server project:
    https://gitlab.esrf.fr/icat/icat-plus/-/blob/master/README.md
    """

    DEFAULT_SCHEME = "https"

    def __init__(
        self,
        url: str,
        api_key: str = "elogbook-00000000-0000-0000-0000-000000000000",
        timeout: Optional[float] = None,
        **payload,
    ):
        url = normalize_url(url, default_scheme=self.DEFAULT_SCHEME)

        path = f"dataacquisition/{api_key}/notification"
        self._message_url = urljoin(url, path)

        path = f"dataacquisition/{api_key}/base64"
        self._data_url = urljoin(url, path)

        self._init_payload = payload
        self._init_payload.setdefault("machine", socket.getfqdn())
        self._init_payload.setdefault("software", "pyicat-plus_v" + __version__)

        self.raise_error = True
        if timeout is None:
            timeout = 0.1
        self.timeout = timeout

    def _merge_payloads(self, message_payload: dict, call_payload: dict) -> dict:
        payloads = self._sorted_payloads(message_payload, call_payload)
        result = {k: v for payload in payloads for k, v in payload.items()}
        tags = self._merge_payload_tags(*payloads)
        if tags:
            result.pop("tags", None)
            result["tag"] = tags
        return result

    def _sorted_payloads(self, message_payload: dict, call_payload: dict) -> List[dict]:
        """Sorted by increasing priority"""
        return [message_payload, self._init_payload, call_payload]

    def _merge_payload_tags(self, *payloads: Iterable[dict]) -> List[dict]:
        """The payload tags can be eithers a list of strings or a list of dictionaries.
        The return value are the merged tags as a list of dictionaries.
        """
        names = set()
        tags = list()
        for payload in payloads:
            ptags = payload.get("tag", list()) + payload.get("tags", list())
            for tag in ptags:
                if isinstance(tag, str):
                    if tag in names:
                        continue
                    names.add(tag)
                    tags.append({"name": tag})
                else:
                    if tag["name"] in names:
                        continue
                    names.add(tag["name"])
                    tags.append(tag)
        return tags

    def _post_with_payload(
        self, url: str, message_payload: dict, call_payload: dict
    ) -> None:
        payload = self._merge_payloads(message_payload, call_payload)
        payload.setdefault("creationDate", datetime.now().isoformat())
        try:
            response = requests.post(url, json=payload, timeout=self.timeout)
        except requests.exceptions.ReadTimeout:
            return  # we have no confirmation that the call succeeded
        except Exception as e:
            if self.raise_error:
                raise
            logger.exception(e)
            return
        if self.raise_error:
            response.raise_for_status()
        elif not response.ok:
            logger.error("%s: %s", response, response.text)

    def send_message(
        self,
        message: str,
        category: str,
        beamline: Optional[str] = None,
        proposal: Optional[str] = None,
        dataset: Optional[str] = None,
        **call_payload,
    ):
        url = self._compose_url(
            url=self._message_url, beamline=beamline, proposal=proposal
        )
        message_payload = self._encode_message(message, category, dataset=dataset)
        self._post_with_payload(url, message_payload, call_payload)

    def send_binary_data(
        self,
        data: bytes,
        mimetype: str,
        beamline: Optional[str] = None,
        proposal: Optional[str] = None,
        **call_payload,
    ):
        url = self._compose_url(
            url=self._data_url, beamline=beamline, proposal=proposal
        )
        message_payload = self._encode_binary_data(data, mimetype=mimetype)
        self._post_with_payload(url, message_payload, call_payload)

    @staticmethod
    def _compose_url(
        url: str, beamline: Optional[str] = None, proposal: Optional[str] = None
    ):
        query = {}
        if beamline:
            query["instrumentName"] = beamline
        if proposal:
            query["investigationName"] = proposal
        query = "&".join([f"{k}={v}" for k, v in query.items()])
        return f"{url}?{query}"

    def send_text_file(
        self,
        filename: str,
        beamline: Optional[str] = None,
        proposal: Optional[str] = None,
        dataset: Optional[str] = None,
        **payload,
    ):
        with open(filename, "r") as f:
            message = f.read()
        self.send_message(
            message,
            category="comment",
            proposal=proposal,
            beamline=beamline,
            dataset=dataset,
            **payload,
        )

    def send_binary_file(
        self,
        filename: str,
        beamline: Optional[str] = None,
        proposal: Optional[str] = None,
        **payload,
    ):
        with open(filename, "rb") as f:
            data = f.read()
        mimetype, _ = mimetypes.guess_type(filename, strict=True)
        self.send_binary_data(
            data, mimetype=mimetype, beamline=beamline, proposal=proposal, **payload
        )

    def _encode_message(
        self,
        message: str,
        category: str,
        dataset: Optional[str] = None,
    ) -> dict:
        try:
            category = MessageCategoryMapping[category.lower()]
        except KeyError:
            raise ValueError(category, "Not a valid e-logbook category") from None
        message_type = MessageTypeMapping[category]
        message = {
            "type": message_type.name,
            "category": category.name,
            "content": [{"format": "plainText", "text": message}],
        }
        if dataset:
            message["datasetName"] = dataset
        return message

    def _encode_binary_data(
        self,
        data: bytes,
        mimetype: Optional[str] = None,
    ) -> dict:
        if not mimetype:
            # arbitrary binary data
            mimetype = "application/octet-stream"
        data_header = f"data:{mimetype};base64,"
        data_blob = base64.b64encode(data).decode("latin-1")
        return {"base64": data_header + data_blob}



from typing import Optional, List, Tuple, Sequence
from dataclasses import dataclass


@dataclass(frozen=True)
class DatasetId:
    name: str
    path: str


class IcatClientInterface:
    def send_message(
        self,
        msg: str,
        msg_type="comment",
        beamline: Optional[str] = None,
        proposal: Optional[str] = None,
        dataset: Optional[str] = None,
        beamline_only: Optional[bool] = None,
        **payload
    ):
        raise NotImplementedError

    def send_data(
        self,
        data: bytes,
        mimetype: Optional[str] = None,
        beamline: Optional[str] = None,
        proposal: Optional[str] = None,
        beamline_only: Optional[bool] = None,
        **payload
    ):
        raise NotImplementedError

    def send_text_file(
        self,
        filename: str,
        beamline: Optional[str] = None,
        proposal: Optional[str] = None,
        dataset: Optional[str] = None,
        beamline_only: Optional[bool] = None,
        **payload
    ):
        raise NotImplementedError

    def send_binary_file(
        self,
        filename: str,
        beamline: Optional[str] = None,
        proposal: Optional[str] = None,
        beamline_only: Optional[bool] = None,
        **payload
    ):
        raise NotImplementedError

    def start_investigation(
        self,
        beamline: Optional[str] = None,
        proposal: Optional[str] = None,
        start_datetime=None,
    ):
        raise NotImplementedError

    def store_dataset(
        self,
        beamline: Optional[str] = None,
        proposal: Optional[str] = None,
        dataset: Optional[str] = None,
        path: Optional[str] = None,
        metadata: dict = None,
        store_filename: Optional[str] = None,
    ):
        raise NotImplementedError

    def store_processed_dataset(
        self,
        beamline: Optional[str] = None,
        proposal: Optional[str] = None,
        dataset: Optional[str] = None,
        path: Optional[str] = None,
        metadata: dict = None,
        raw: Sequence = tuple(),
        store_filename: Optional[str] = None,
    ):
        raise NotImplementedError

    def store_dataset_from_file(self, store_filename: Optional[str] = None):
        raise NotImplementedError

    def investigation_info(
        self, beamline: str, proposal: str, timeout: Optional[float] = None
    ) -> Optional[dict]:
        raise NotImplementedError

    def registered_dataset_ids(
        self, beamline: str, proposal: str
    ) -> Optional[List[DatasetId]]:
        raise NotImplementedError

    def investigation_info_string(
        self, beamline: str, proposal: str, timeout: Optional[float] = None
    ) -> str:
        raise NotImplementedError

    def investigation_summary(
        self, beamline: str, proposal: str, timeout: Optional[float] = None
    ) -> List[Tuple]:
        raise NotImplementedError

    @property
    def expire_datasets_on_close(self) -> bool:
        raise NotImplementedError

    @property
    def reason_for_missing_information(self) -> str:
        raise NotImplementedError



from functools import partial
from typing import Callable, Iterable, Optional


class NamespaceWrapper(object):
    """Namespace which delegates attribute getting and setting to a getter and setter method."""

    def __new__(
        cls,
        property_names: Iterable[str],
        getter: Callable,
        setter: Optional[Callable] = None,
        property_decorator: Optional[Callable] = None,
    ):
        cls = type(cls.__name__, (cls,), {})
        if property_decorator is None:
            property_decorator = property
        for key in property_names:
            prop = property_decorator(partial(NamespaceWrapper._getter, key=key))
            if setter is not None:
                prop = prop.setter(partial(NamespaceWrapper._setter, key=key))
            setattr(cls, key, prop)
        return object.__new__(cls)

    def __init__(
        self,
        property_names: Iterable[str],
        getter: Callable,
        setter: Optional[Callable] = None,
        property_decorator: Optional[Callable] = None,
    ):
        self.__property_names = property_names
        self.__getter = getter
        self.__setter = setter

    def _getter(self, key):
        return self.__getter(key)

    def _setter(self, value, key):
        return self.__setter(key, value)

    def get_content(self) -> str:
        if not self.__property_names:
            return "Namespace is empty"
        res = "Namespace contains:\n"
        max_len = max(len(s) for s in self.__property_names)
        key_fmt = f".%-{max_len}s"
        for key in self.__property_names:
            val = self._getter(key)
            if val:
                res += (key_fmt % key) + f" = {val!r}\n"
            else:
                res += (key_fmt % key) + "\n"
        return res



import os
import re
import requests
from enum import Enum
from io import StringIO
from dataclasses import dataclass
from collections.abc import Mapping, Sequence
from typing import Any, Callable, Dict, Iterable, Iterator, Optional, Union
from xml.etree import ElementTree

from .namespace_wrapper import NamespaceWrapper


IcatNodeIdLike = Union[str, Iterable[str]]
IcatItemType = Union["IcatField", "IcatFieldGroup"]
IcatCategory = Enum(
    "IcatCategory",
    "other technique instrument sample positioner note slit attenuator detector",
)
NX_CLASS_TO_CATEGORY = {
    "NXsubentry": IcatCategory.technique,
    "NXpositioner": IcatCategory.positioner,
    "NXinstrument": IcatCategory.instrument,
    "NXsample": IcatCategory.sample,
    "NXnote": IcatCategory.note,
    "NXslit": IcatCategory.slit,
    "NXattenuator": IcatCategory.attenuator,
    "NXdetector": IcatCategory.detector,
}


class IcatNodeId(Sequence):
    """Identifier of an item in the ICAT metadata tree."""

    def __init__(self, iterable: Optional[IcatNodeIdLike] = None) -> None:
        self._sequence = tuple(self._parse_init(iterable))

    def _parse_init(self, iterable: Optional[IcatNodeIdLike] = None) -> Iterable[str]:
        if iterable is None:
            return
        if isinstance(iterable, str):
            yield iterable
            return
        for s in iterable:
            if not isinstance(s, str):
                raise TypeError(s)
            if s:
                yield s

    def __getitem__(self, idx) -> str:
        value = self._sequence[idx]
        if isinstance(value, tuple):
            return type(self)(value)
        return value

    def __len__(self) -> int:
        return len(self._sequence)

    def __str__(self) -> str:
        return ".".join(self)

    def __repr__(self) -> str:
        return repr(str(self))

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, type(self)):
            return False
        return self._sequence == other._sequence

    def __add__(self, other: IcatNodeIdLike):
        if isinstance(other, str):
            other = (other,)
        elif not isinstance(other, tuple):
            other = tuple(other)
        return type(self)(self._sequence + other)

    def endswith(self, suffix: IcatNodeIdLike):
        if isinstance(suffix, str):
            suffix = (suffix,)
        elif not isinstance(suffix, tuple):
            suffix = tuple(suffix)
        return self._sequence[-len(suffix) :] == suffix


def as_icatnodeid(node_id: IcatNodeIdLike) -> IcatNodeId:
    if isinstance(node_id, IcatNodeId):
        return node_id
    else:
        return IcatNodeId(node_id)


@dataclass(frozen=True, repr=True, eq=True, order=True)
class IcatField:
    """Description of a single ICAT database field"""

    name: str
    field_name: str
    parent: IcatNodeId
    nxtype: str
    description: Optional[str]
    units: Optional[str]

    @property
    def node_id(self) -> IcatNodeId:
        """Node identifier in the metadata tree"""
        return self.parent + self.name

    @property
    def info(self) -> "IcatField":
        return self


@dataclass(frozen=True, repr=True, eq=True, order=True)
class IcatGroup:
    """Description of a group of ICAT database items"""

    name: str
    parent: IcatNodeId
    NX_class: str

    @property
    def node_id(self) -> IcatNodeId:
        """Node identifier in the metadata tree"""
        return self.parent + self.name

    @property
    def category(self) -> IcatCategory:
        return NX_CLASS_TO_CATEGORY.get(self.NX_class, IcatCategory.other)

    @property
    def info(self) -> "IcatGroup":
        return self


class IcatFieldGroup(Mapping):
    """A group of ICAT database items"""

    def __init__(self, info: IcatGroup, nodes: Dict[str, IcatItemType]) -> None:
        self.__info = info
        self.__nodes = nodes

    def __getitem__(self, node_id: IcatNodeIdLike) -> IcatItemType:
        node_id = as_icatnodeid(node_id)
        adict = self.__nodes
        for name in node_id:
            adict = adict[name]
        return adict

    def __iter__(self) -> Iterator[str]:
        return self.__nodes.__iter__()

    def __len__(self) -> int:
        return self.__nodes.__len__()

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self.info!r})"

    @property
    def info(self) -> IcatGroup:
        return self.__info

    def iter_fields(self) -> Iterable[IcatField]:
        for icat_item in self.values():
            if isinstance(icat_item, IcatField):
                yield icat_item
            else:
                yield from icat_item.iter_fields()

    def iter_groups(self) -> Iterable["IcatFieldGroup"]:
        cls = type(self)
        for icat_item in self.values():
            if isinstance(icat_item, cls):
                yield icat_item
                yield from icat_item.iter_groups()

    def iter_group_names(self) -> Iterable[str]:
        for group in self.iter_groups():
            yield str(group.info.node_id)

    def iter_field_names(self) -> Iterable[str]:
        for field in self.iter_fields():
            yield str(field.node_id)

    def get_field_with_field_name(self, field_name: str) -> Optional[IcatField]:
        for field in self.iter_fields():
            if field.field_name == field_name:
                return field

    def iter_groups_with_type(
        self, categories: Union[IcatCategory, str, Iterable[Union[IcatCategory, str]]]
    ) -> Iterable["IcatFieldGroup"]:
        if isinstance(categories, (IcatCategory, str)):
            categories = {categories}
        categories = {
            category
            if isinstance(category, IcatCategory)
            else IcatCategory.__members__[category]
            for category in categories
        }
        for group in self.iter_groups():
            if group.info.category in categories:
                yield group

    def iter_items_with_node_id_suffix(
        self, node_id_suffix: IcatNodeIdLike
    ) -> Iterable["IcatFieldGroup"]:
        node_id_suffix = as_icatnodeid(node_id_suffix)
        cls = type(self)
        for icat_item in self.values():
            if icat_item.info.node_id.endswith(node_id_suffix):
                yield icat_item
            elif isinstance(icat_item, cls):
                yield from icat_item.iter_items_with_node_id_suffix(node_id_suffix)

    def namespace(
        self,
        getter: Optional[Callable[[Any, str], Any]] = None,
        setter: Optional[Callable[[Any, str, Any], None]] = None,
        property_decorator: Optional[Callable] = None,
    ) -> NamespaceWrapper:
        adict = dict()

        if setter is None:
            wrap_setter = None
        else:

            def wrap_setter(key, value):
                field = adict.get(key, None)
                if isinstance(field, NamespaceWrapper):
                    raise AttributeError(f"'{key}' is not an ICAT field")
                setter(field.field_name, value)

        def wrap_getter(key):
            value = adict[key]
            if getter is None:
                return value
            elif isinstance(value, NamespaceWrapper):
                return value
            else:
                return getter(value.info.field_name)

        cls = type(self)
        for k, v in self.items():
            if isinstance(v, cls):
                adict[k] = v.namespace(getter=getter, setter=setter)
            else:
                adict[k] = v
        return NamespaceWrapper(
            property_names=list(adict),
            property_decorator=property_decorator,
            getter=wrap_getter,
            setter=wrap_setter,
        )

    def valid_field_name(self, field_name: str) -> bool:
        for field in self.iter_fields():
            if field.field_name == field_name:
                return True
        return False


ICAT_FIELD_NAME = re.compile(r"^\$\{(.+)\}$")


def load_group(in_node: ElementTree.Element, **node_attributes) -> IcatFieldGroup:
    """Convert an XML node to an `IcatFieldGroup` instance"""
    fields = list()
    groups = list()
    for element in in_node:
        if element.text is None:
            match = None
            assert element.tag == "link"
            continue
        match = ICAT_FIELD_NAME.match(element.text)
        if match:
            assert element.tag != "group"
            field_name = match.groups()[0]
            fields.append((element.tag, field_name, element.attrib))
        else:
            assert element.tag == "group"
            group_name = element.attrib["groupName"]
            groups.append((group_name, element))

    info = IcatGroup(**node_attributes)
    parent = info.node_id
    nodes = dict()

    for name, field_name, attrs in fields:
        icat_item = IcatField(
            name=name,
            field_name=field_name,
            parent=parent,
            nxtype=attrs["NAPItype"],
            description=attrs.get("ESRF_description"),
            units=attrs.get("units"),
        )
        nodes[icat_item.name] = icat_item
    for group_name, element in groups:
        icat_item = load_group(
            element, name=group_name, parent=parent, NX_class=element.attrib["NX_class"]
        )
        nodes[icat_item.info.name] = icat_item

    return IcatFieldGroup(info, nodes)


def icat_fields_source(url: Optional[str] = None) -> Union[str, StringIO]:
    """Get XML description of all ICAT fields from a URL (local file by default)"""
    if not url:
        return os.path.join(os.path.dirname(__file__), "hdf5_cfg.xml")
    r = requests.get(url)
    return StringIO(r.text)


def load_icat_fields(url: Optional[str] = None) -> IcatFieldGroup:
    """Returns an object which allows browsing the ICAT metadata definitions.

    Example URL: "https://gitlab.esrf.fr/icat/hdf5-master-config/-/raw/master/hdf5_cfg.xml"
    """
    tree = ElementTree.parse(icat_fields_source(url=url))
    in_node = tree.getroot()
    return load_group(in_node, name="", parent=IcatNodeId(), NX_class="NXentry")



from ..metadata.definitions import load_icat_fields
from ..metadata.definitions import IcatNodeId


def test_getitem():
    icat_fields = load_icat_fields()
    for field in icat_fields.iter_fields():
        assert icat_fields[field.node_id] == field


def test_node_id():
    assert IcatNodeId()[:] == IcatNodeId()
    assert IcatNodeId("a")[0] == "a"
    assert IcatNodeId(("a", "b", "c"))[:2] == IcatNodeId(("a", "b"))
    assert IcatNodeId(("a", "b", "c")).endswith(("b", "c"))


def test_find_groups_by_item_id():
    icat_fields = load_icat_fields()
    groups = list(icat_fields.iter_items_with_node_id_suffix("primary_slit"))
    assert len(groups) == 1
    assert str(groups[0].info.node_id) == "instrument.primary_slit"



def test_stomp(icat_publisher, icat_subscriber):
    icat_publisher.sendall(b"MYMESSAGE1\nMYMESSAGE2\n")
    assert icat_subscriber.get(timeout=5) == "MYMESSAGE1"
    assert icat_subscriber.get(timeout=5) == "MYMESSAGE2"


def test_activemq_rest_server(activemq_rest_server):
    # TODO: send test request
    pass


def test_icat_logbook_server(icat_logbook_subscriber):
    # TODO: send test request
    pass



import os
import re
import socket
import json
import http.server
import logging
from datetime import datetime

from pyicat_plus.tests.servers.utils import ReuseAddrTCPServer
from pyicat_plus.tests.servers.utils import basic_config

logger = logging.getLogger("ICATPLUS SERVER")
basic_config(
    logger=logger,
    level=logging.DEBUG,
    format="%(asctime)s - %(name)s - %(levelname)s - %(message)s",
)


class MyTCPRequestHandler(http.server.BaseHTTPRequestHandler):
    def __init__(self, *args, s_out=None, icat_data_dir=None, **kw):
        self.s_out = s_out
        self.icat_data_dir = icat_data_dir
        self._investigation_id = 1234
        self._investigation_startdate = datetime.now().astimezone().isoformat()
        super().__init__(*args, **kw)

    def do_HEAD(self):
        logger.info("HEAD")
        self.reply_ok()

    def do_GET(self):
        if self._handle_investigation():
            return
        if self._handle_dataset():
            return
        logger.error("Bad GET request: %s", self.path)
        self.reply_bad_request()

    def do_POST(self):
        if self._handle_elogbook_message():
            return
        if self._handle_elogbook_data():
            return
        if self._handle_elogbook_message_beamline_only():
            return
        if self._handle_elogbook_message_no_beamline():
            return
        if self._handle_elogbook_data_beamline_only():
            return
        logger.error("Bad POST request: %s", self.path)
        self.reply_bad_request()

    def _handle_investigation(self):
        fmt = "/dataacquisition/(?P<apikey>[^//]+)/investigation\\?instrumentName=(?P<instrument>[^//]+)&investigationName=(?P<investigation>[^//]+)&sortBy=startdate"
        mresult = re.match(fmt, self.path)
        if not mresult:
            return
        if "666" in mresult.groupdict()["investigation"]:
            self.reply_error()
            return True
        response = [
            {"id": self._investigation_id, "startDate": self._investigation_startdate}
        ]
        logger.info("Investigation GET response: %s", response)
        self.reply_json(response)
        return True

    def _handle_dataset(self):
        fmt = f"/dataacquisition/(?P<apikey>[^//]+)/dataset\\?investigationId={self._investigation_id}"
        mresult = re.match(fmt, self.path)
        if not mresult:
            return
        datasets = list()
        if self.icat_data_dir:
            for name in os.listdir(self.icat_data_dir):
                filename = os.path.join(self.icat_data_dir, name)
                with open(filename, "r") as content:
                    datasets.append(json.load(content))
        logger.info("Dataset GET response: %s", datasets)
        self.reply_json(datasets)
        return True

    def _handle_elogbook_message(self):
        logger.info(self.path)
        fmt = "/dataacquisition/(?P<apikey>[^//]+)/notification\\?instrumentName=(?P<instrument>[^//]+)&investigationName=(?P<investigation>[^//]+)"
        mresult = re.match(fmt, self.path)
        if not mresult:
            return
        if (
            self.headers.get("content-type") != "application/json"
            or "666" in mresult.groupdict()["investigation"]
        ):
            self.reply_bad_request()
            return True
        length = int(self.headers.get("content-length"))
        adict = json.loads(self.rfile.read(length))
        adict.update(mresult.groupdict())
        self.on_elog_message(adict)
        self.reply_ok()
        return True

    def _handle_elogbook_data(self):
        fmt = "/dataacquisition/(?P<apikey>[^//]+)/base64\\?instrumentName=(?P<instrument>[^//]+)&investigationName=(?P<investigation>[^//]+)"
        mresult = re.match(fmt, self.path)
        if not mresult:
            return
        if (
            self.headers.get("content-type") != "application/json"
            or "666" in mresult.groupdict()["investigation"]
        ):
            self.reply_bad_request()
            return True
        length = int(self.headers.get("content-length"))
        adict = json.loads(self.rfile.read(length))
        adict.update(mresult.groupdict())
        self.on_elog_message(adict)
        self.reply_ok()
        return True

    def _handle_elogbook_message_beamline_only(self):
        logger.info(self.path)
        fmt = "/dataacquisition/(?P<apikey>[^//]+)/notification\\?instrumentName=(?P<instrument>[^//]+)"
        mresult = re.match(fmt, self.path)
        if not mresult:
            return
        if self.headers.get("content-type") != "application/json":
            self.reply_bad_request()
            return True
        length = int(self.headers.get("content-length"))
        adict = json.loads(self.rfile.read(length))
        adict.update(mresult.groupdict())
        self.on_elog_message(adict)
        self.reply_ok()
        return True

    def _handle_elogbook_message_no_beamline(self):
        logger.info(self.path)
        fmt = "/dataacquisition/(?P<apikey>[^//]+)/notification"
        mresult = re.match(fmt, self.path)
        if not mresult:
            return
        if self.headers.get("content-type") != "application/json":
            self.reply_bad_request()
            return True
        length = int(self.headers.get("content-length"))
        adict = json.loads(self.rfile.read(length))
        adict.update(mresult.groupdict())
        self.on_elog_message(adict)
        self.reply_ok()
        return True

    def _handle_elogbook_data_beamline_only(self):
        fmt = "/dataacquisition/(?P<apikey>[^//]+)/base64\\?instrumentName=(?P<instrument>[^//]+)"
        mresult = re.match(fmt, self.path)
        if not mresult:
            return
        if self.headers.get("content-type") != "application/json":
            self.reply_bad_request()
            return True
        length = int(self.headers.get("content-length"))
        adict = json.loads(self.rfile.read(length))
        adict.update(mresult.groupdict())
        self.on_elog_message(adict)
        self.reply_ok()
        return True

    def reply_ok(self):
        self.send_response(http.HTTPStatus.OK)
        self.end_headers()

    def reply_bad_request(self):
        self.send_response(http.HTTPStatus.BAD_REQUEST)
        self.end_headers()

    def reply_not_found(self):
        self.send_response(http.HTTPStatus.NOT_FOUND)
        self.end_headers()

    def reply_error(self):
        self.send_response(http.HTTPStatus.INTERNAL_SERVER_ERROR)
        self.end_headers()

    def reply_json(self, adict):
        body = json.dumps(adict).encode()
        self.send_response(http.HTTPStatus.OK)
        self.send_header("Content-type", "application/json")
        self.send_header("Content-Length", str(len(body)))
        self.end_headers()
        self.wfile.write(body)

    def on_elog_message(self, adict):
        if not adict:
            return
        logger.info("received message:\n %s", adict)
        if self.s_out is not None:
            self.s_out.sendall(json.dumps(adict).encode() + b"\n")


def main(port=8443, port_out=0, icat_data_dir=None):
    if port_out:
        s_out = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s_out.connect(("localhost", port_out))
    else:
        s_out = None

    class MyTCPRequestHandlerWithSocket(MyTCPRequestHandler):
        def __init__(self, *args, **kw):
            super().__init__(*args, s_out=s_out, icat_data_dir=icat_data_dir, **kw)

    # Create a TCP Server instance
    aServer = ReuseAddrTCPServer(("localhost", port), MyTCPRequestHandlerWithSocket)

    # Start accepting requests and setup output socket
    if port_out:
        logger.info(f"Redirect received messages to port {port_out}")
        s_out.sendall(json.dumps({"STATUS": "LISTENING"}).encode() + b"\n")
    try:
        logger.info("CTRL-C to stop")
        aServer.serve_forever()
    finally:
        logger.info("Exit.")


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser(description="ICAT plus server")
    parser.add_argument("--port", default=8443, type=int, help="server port")
    parser.add_argument("--port_out", default=0, type=int, help="output socket")
    parser.add_argument(
        "--icat_data_dir", default="", type=str, help="One file per dataset"
    )
    args = parser.parse_args()
    main(port=args.port, port_out=args.port_out, icat_data_dir=args.icat_data_dir)



from typing import List, Tuple
from .interface import IcatClientInterface


class IcatNullClient(IcatClientInterface):
    def __init__(self, expire_datasets_on_close=True) -> None:
        self.__expire_datasets_on_close = expire_datasets_on_close

    def send_message(self, *args, **kw):
        pass

    def send_data(self, *args, **kw):
        pass

    def send_text_file(self, *args, **kw):
        pass

    def send_binary_file(self, *args, **kw):
        pass

    def start_investigation(self, *args, **kw):
        pass

    def store_dataset(self, *args, **kw):
        pass

    def store_processed_dataset(self, *args, **kw):
        pass

    def store_dataset_from_file(self, *args, **kw):
        pass

    def investigation_info(self, *args, **kwargs) -> None:
        pass

    def registered_dataset_ids(self, *args, **kwargs) -> None:
        pass

    def investigation_info_string(self, *args, **kwargs) -> str:
        return f"Proposal information not available ({self.reason_for_missing_information})"

    def investigation_summary(self, *args, **kwargs) -> List[Tuple]:
        keys = ["e-logbook", "data portal"]
        return [
            (key, f"information not available ({self.reason_for_missing_information})")
            for key in keys
        ]

    @property
    def expire_datasets_on_close(self) -> bool:
        return self.__expire_datasets_on_close

    @property
    def reason_for_missing_information(self) -> str:
        if self.__expire_datasets_on_close:
            return "ICAT is not configured"
        else:
            return "ICAT communication is disabled"



import time
import json
import socket
from contextlib import contextmanager
from .misc import eprint
from ...concurrency import gevent, Queue, spawn


def get_open_port():
    s = socket.socket()
    try:
        s.bind(("", 0))
        return s.getsockname()[1]
    finally:
        s.close()


def wait_tcp_online(host, port, timeout=5):
    """Wait for a TCP port with a timeout.

    Raises a `gevent.Timeout` if the port was not found.
    """
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    sock.settimeout(timeout)
    try:
        while True:
            try:
                sock.connect((host, port))
                break
            except ConnectionError:
                pass
    finally:
        sock.close()


@contextmanager
def tcp_message_server(data_parser=None, validate_all=True, timeout=5):
    """Start a TCP server and yield a queue of events.
    Data packages are separated by newline characters.
    Supported package encodings are UTF8 (default) and json.
    """
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    sock.settimeout(timeout)
    port = get_open_port()
    sock.bind(("localhost", port))
    sock.listen()

    # Listen to this socket
    messages = Queue()
    stop = False

    def listener():
        buffer = b""
        while True:
            try:
                conn, addr = sock.accept()
                break
            except socket.timeout:
                time.sleep(0.1)
        while True:
            try:
                buffer += conn.recv(16384)
            except socket.timeout:
                pass
            if buffer:
                out, sep, buffer = buffer.rpartition(b"\n")
                if sep:
                    for bdata in out.split(b"\n"):
                        if data_parser == "json":
                            messages.put(json.loads(bdata))
                        else:
                            messages.put(bdata.decode())
            if stop:
                return
            time.sleep(0.1)

    glistener = spawn(listener)
    try:
        yield port, messages
    finally:
        messages.put(StopIteration)
        stop = True
        if validate_all:
            for msg in iter(messages.get, StopIteration):
                eprint(f"Unvalidated message: {msg}")
        if gevent:
            glistener.kill()
        else:
            glistener.join()
        sock.close()



import os
import re
import json
import stomp
import socket
import logging
import threading

from pyicat_plus.tests.servers.utils import basic_config

logger = logging.getLogger("STOMP SUBSCRIBER")
basic_config(
    logger=logger,
    level=logging.DEBUG,
    format="%(asctime)s - %(name)s - %(levelname)s - %(message)s",
)


class MyListener(stomp.ConnectionListener):
    def __init__(self, conn, icat_data_dir=""):
        self.conn = conn
        self.s_out = None
        self.icat_data_dir = icat_data_dir
        super().__init__()

    def redirect_messages(self, port):
        if self.s_out is not None:
            self.s_out.close()
        self.s_out = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.s_out.connect(("localhost", port))
        logger.info(f"Redirect received messages to port {port}")

    def on_message(self, frame):
        message = frame.body
        header = frame.headers
        if header.get("destination") not in [
            "/queue/icatIngest",
            "/queue/icatArchiveRestoreStatus",
        ]:
            return
        logger.info(f"received message:\n {message}")

        invalid_proposal = re.findall(
            "<tns:experiment>([^<>]*666[^<>]*)<\\/tns:experiment>", message
        ) or re.findall(
            "<tns:investigation>([^<>]*666[^<>]*)<\\/tns:investigation>", message
        )
        if invalid_proposal:
            logger.info(
                "Do not register message for invalid proposal '%s'", invalid_proposal[0]
            )
            return

        if self.s_out is not None:
            self.s_out.sendall(message.encode() + b"\n")

        filename = None
        if self.icat_data_dir:
            fmt = r"\<tns\:location\>(?P<location>.+?)\<\/tns\:location\>"
            results = re.findall(fmt, message)
            if results:
                info = {"location": results[0]}
                name = "{:04d}.json".format(len(os.listdir(self.icat_data_dir)))
                filename = os.path.join(self.icat_data_dir, name)
                with open(filename, "w") as f:
                    json.dump(info, f)
        logger.info(f"dataset file saved: {filename}")


def main(host=None, port=60001, queue=None, port_out=0, icat_data_dir=""):
    if not host:
        host = "localhost"
    if not queue:
        queue = "/queue/icatIngest"
    conn = stomp.Connection([(host, port)])
    # Listener will run in a different thread
    listener = MyListener(conn, icat_data_dir)
    conn.set_listener("", listener)
    conn.connect("guest", "guest", wait=True)
    conn.subscribe(destination=queue, id=1, ack="auto")
    logger.info(f"subscribed to {queue} on STOMP {host}:{port}")
    if port_out:
        listener.redirect_messages(port_out)
        listener.s_out.sendall(b"LISTENING\n")
    logger.info("CTRL-C to stop")
    try:
        threading.Event().wait()
    finally:
        logger.info("Exit.")


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser(
        description="STOMP client which subscribes to a STOMP queue and redirect its output to a socket"
    )
    parser.add_argument(
        "--host", default="localhost", type=str, help="STOMP server host"
    )
    parser.add_argument("--port", default=60001, type=int, help="STOMP server port")
    parser.add_argument(
        "--queue", default="/queue/icatIngest", type=str, help="STOMP queue"
    )
    parser.add_argument("--port_out", default=0, type=int, help="output socket")
    parser.add_argument(
        "--icat_data_dir", default="", type=str, help="Dataset directory"
    )
    args = parser.parse_args()

    main(
        host=args.host,
        port=args.port,
        port_out=args.port_out,
        queue=args.queue,
        icat_data_dir=args.icat_data_dir,
    )



import socket
import logging
from stompest.config import StompConfig
from stompest.protocol import StompSpec
from stompest.sync import Stomp

from pyicat_plus.tests.servers.utils import basic_config

logger = logging.getLogger("STOMP PUBLISHER")
basic_config(
    logger=logger,
    level=logging.DEBUG,
    format="%(asctime)s - %(name)s - %(levelname)s - %(message)s",
)


def read_socket(host, port):
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    sock.connect((host, port))
    buffer = b""
    try:
        while True:
            buffer += sock.recv(16384)
            if buffer:
                out, sep, buffer = buffer.rpartition(b"\n")
                if sep:
                    for bdata in out.split(b"\n"):
                        yield bdata
    finally:
        sock.close()


def main(host=None, port=60001, queue=None, port_in=0):
    if not host:
        host = "localhost"
    if not queue:
        queue = "/queue/icatIngest"

    queueURLs = [f"{host}:{port}"]
    cfgURL = "failover:(tcp://" + ",tcp://".join(queueURLs) + ")"
    cfgURL += "?maxReconnectAttempts=3,initialReconnectDelay=250,maxReconnectDelay=1000"
    client = Stomp(StompConfig(cfgURL, version=StompSpec.VERSION_1_1))
    client.connect(
        versions=[StompSpec.VERSION_1_1], heartBeats=(0, 0), connectedTimeout=1
    )
    header = {
        "persistent": "true",
        StompSpec.ACK_HEADER: StompSpec.ACK_CLIENT_INDIVIDUAL,
    }
    if port_in:
        for body in read_socket("localhost", port_in):
            client.send(queue, body=body, headers=header)
    else:
        while True:
            body = input(f"Message to send to '{queue}': ")
            client.send(queue, body=body.encode(), headers=header)


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser(
        description="Redirect socket input to a STOMP queue"
    )
    parser.add_argument(
        "--host", default="localhost", type=str, help="STOMP server host"
    )
    parser.add_argument("--port", default=60001, type=int, help="STOMP server port")
    parser.add_argument(
        "--queue", default="/queue/icatIngest", type=str, help="STOMP queue"
    )
    parser.add_argument("--port_in", default=0, type=int, help="input socket")
    args = parser.parse_args()

    main(host=args.host, port=args.port, port_in=args.port_in, queue=args.queue)
